import warnings
import numpy as np
import scipy.optimize as spopt
from scipy.constants import hbar

from .utilities import plotting, save_load, Watt2dBm, dBm2Watt
from .circlefit import circlefit
from .calibration import calibration

##
## z_data_raw denotes the raw data
## z_data denotes the normalized data
##        
    
class reflection_port(circlefit, save_load, plotting, calibration):
    '''
    normal direct port probed in reflection
    '''
    def __init__(self, f_data=None, z_data_raw=None):
        self.porttype = 'direct'
        self.fitresults = {}
        self.z_data = None
        if f_data is not None:
            self.f_data = np.array(f_data)
        else:
            self.f_data=None
        if z_data_raw is not None:
            self.z_data_raw = np.array(z_data_raw)
        else:
            self.z_data=None
    
    def _S11(self,f,fr,k_c,k_i):
        '''
        use either frequency or angular frequency units
        for all quantities
        k_l=k_c+k_i: total (loaded) coupling rate
        k_c: coupling rate
        k_i: internal loss rate
        '''
        return ((k_c-k_i)+2j*(f-fr))/((k_c+k_i)-2j*(f-fr))
    
    def get_delay(self,f_data,z_data,delay=None,ignoreslope=True,guess=True):
        '''
        ignoreslope option not used here
        retrieves the cable delay assuming the ideal resonance has a circular shape
        modifies the cable delay until the shape Im(S21) vs Re(S21) is circular
        see "do_calibration"
        '''
        maxval = np.max(np.absolute(z_data))
        z_data = z_data/maxval
        A1, A2, A3, A4, fr, Ql = self._fit_skewed_lorentzian(f_data,z_data)
        if ignoreslope==True:
            A2 = 0
        else:
            z_data = (np.sqrt(np.absolute(z_data)**2-A2*(f_data-fr))) * np.exp(np.angle(z_data)*1j)  #usually not necessary
        if delay==None:
            if guess==True:
                delay = self._guess_delay(f_data,z_data)
            else:
                delay=0.
            delay = self._fit_delay(f_data,z_data,delay,maxiter=200)
        params = [A1, A2, A3, A4, fr, Ql]
        return delay, params 
    
    def do_calibration(self,f_data,z_data,ignoreslope=True,guessdelay=True):
        '''
        calculating parameters for normalization
        '''
        delay, params = self.get_delay(f_data,z_data,ignoreslope=ignoreslope,guess=guessdelay)
        z_data = np.sqrt(np.absolute(z_data)**2-params[1]*(f_data-params[4]))*np.exp(2.*1j*np.pi*delay*f_data)*np.exp(1j*np.angle(z_data))
        xc, yc, r0 = self._fit_circle(z_data)
        zc = np.complex(xc,yc)
        fitparams = self._phase_fit(f_data,self._center(z_data,zc),0.,np.absolute(params[5]),params[4])
        theta, Ql, fr = fitparams
        beta = self._periodic_boundary(theta+np.pi,np.pi) ###
        offrespoint = np.complex((xc+r0*np.cos(beta)),(yc+r0*np.sin(beta)))
        alpha = self._periodic_boundary(np.angle(offrespoint)+np.pi,np.pi)
        #a = np.absolute(offrespoint)
        #alpha = np.angle(zc)
        a = r0 + np.absolute(zc)
        return delay, a, alpha, fr, Ql, params[1], params[4]
    
    def do_normalization(self,f_data,z_data,delay,amp_norm,alpha,A2,frcal):
        '''
        transforming resonator into canonical position
        '''
        return (np.sqrt(np.absolute(z_data)**2-A2*(f_data-frcal)))/amp_norm*np.exp(1j*(-alpha+2.*np.pi*delay*f_data))*np.exp(1j*np.angle(z_data))
    
    def circlefit(self,f_data,z_data,fr=None,Ql=None,refine_results=False,calc_errors=True):
        '''
        S11 version of the circlefit
        '''
    
        if fr==None: fr=f_data[np.argmin(np.absolute(z_data))]
        if Ql==None: Ql=1e6
        xc, yc, r0 = self._fit_circle(z_data,refine_results=refine_results)
        phi0 = -np.arcsin(yc/r0)
        theta0 = self._periodic_boundary(phi0+np.pi,np.pi)
        z_data_corr = self._center(z_data,np.complex(xc,yc))
        theta0, Ql, fr = self._phase_fit(f_data,z_data_corr,theta0,Ql,fr)
        #print("Ql from phasefit is: " + str(Ql))
        Qi = Ql/(1.-r0)
        Qc = 1./(1./Ql-1./Qi)
    
        results = {"Qi":Qi,"Qc":Qc,"Ql":Ql,"fr":fr,"theta0":theta0}
    
        #calculation of the error
        p = [fr,Qc,Ql]
        #chi_square, errors = rt.get_errors(rt.residuals_notch_ideal,f_data,z_data,p)
        if calc_errors==True:
            chi_square, cov = self._get_cov_fast_directrefl(f_data,z_data,p)
            #chi_square, cov = rt.get_cov(rt.residuals_notch_ideal,f_data,z_data,p)
    
            if cov is not None:
                errors = np.sqrt(np.diagonal(cov))
                fr_err,Qc_err,Ql_err = errors
                #calc Qi with error prop (sum the squares of the variances and covariaces)
                dQl = 1./((1./Ql-1./Qc)**2*Ql**2)
                dQc = - 1./((1./Ql-1./Qc)**2*Qc**2)
                Qi_err = np.sqrt((dQl**2*cov[2][2]) + (dQc**2*cov[1][1])+(2*dQl*dQc*cov[2][1]))  #with correlations
                errors = {"Ql_err":Ql_err, "Qc_err":Qc_err, "fr_err":fr_err,"chi_square":chi_square,"Qi_err":Qi_err}
                results.update( errors )
            else:
                print("WARNING: Error calculation failed!")
        else:
            #just calc chisquared:
            fun2 = lambda x: self._residuals_notch_ideal(x,f_data,z_data)**2
            chi_square = 1./float(len(f_data)-len(p)) * (fun2(p)).sum()
            errors = {"chi_square":chi_square}
            results.update(errors)
    
        return results
        
    
    def autofit(self):
        '''
        automatic calibration and fitting
        '''
        delay, amp_norm, alpha, fr, Ql, A2, frcal =\
                self.do_calibration(self.f_data,self.z_data_raw,ignoreslope=True,guessdelay=False)
        self.z_data = self.do_normalization(self.f_data,self.z_data_raw,delay,amp_norm,alpha,A2,frcal)
        self.fitresults = self.circlefit(self.f_data,self.z_data,fr,Ql,refine_results=False,calc_errors=True)
        self.z_data_sim = A2*(self.f_data-frcal)+self._S11_directrefl(self.f_data,fr=self.fitresults["fr"],Ql=self.fitresults["Ql"],Qc=self.fitresults["Qc"],a=amp_norm,alpha=alpha,delay=delay)

    def _S11_directrefl(self,f,fr=10e9,Ql=900,Qc=1000.,a=1.,alpha=0.,delay=.0):
        '''
        full model for notch type resonances
        '''
        return a*np.exp(np.complex(0,alpha))*np.exp(-2j*np.pi*f*delay) * ( 2.*Ql/Qc - 1. + 2j*Ql*(fr-f)/fr ) / ( 1. - 2j*Ql*(fr-f)/fr )    
        
    def get_single_photon_limit(self,unit='dBm'):
        '''
        returns the amout of power in units of W necessary
        to maintain one photon on average in the cavity
        unit can be 'dbm' or 'watt'
        '''
        if self.fitresults!={}:
            fr = self.fitresults['fr']
            k_c = fr/self.fitresults['Qc']
            k_i = fr/self.fitresults['Qi']
            if unit=='dBm':
                return Watt2dBm(1./(4.*k_c/(2.*np.pi*hbar*fr*(k_c+k_i)**2)))
            elif unit=='watt':
                return 1./(4.*k_c/(2.*np.pi*hbar*fr*(k_c+k_i)**2))
                
        else:
            warnings.warn('Please perform the fit first',UserWarning)
            return None
        
    def get_photons_in_resonator(self,power,unit='dBm'):
        '''
        returns the average number of photons
        for a given power (defaul unit is 'dbm')
        unit can be 'dBm' or 'watt'
        '''
        if self.fitresults!={}:
            if unit=='dBm':
                power = dBm2Watt(power)
            fr = self.fitresults['fr']
            k_c = fr/self.fitresults['Qc']
            k_i = fr/self.fitresults['Qi']
            return 4.*k_c/(2.*np.pi*hbar*fr*(k_c+k_i)**2) * power
        else:
            warnings.warn('Please perform the fit first',UserWarning)
            return None
    
class notch_port(circlefit, save_load, plotting, calibration):
    '''
    notch type port probed in transmission
    '''
    def __init__(self, f_data=None, z_data_raw=None):
        self.porttype = 'notch'
        self.fitresults = {}
        self.z_data = None
        if f_data is not None:
            self.f_data = np.array(f_data)
        else:
            self.f_data=None
        if z_data_raw is not None:
            self.z_data_raw = np.array(z_data_raw)
        else:
            self.z_data_raw=None
    
    def get_delay(self,f_data,z_data,delay=None,ignoreslope=True,guess=True):
        '''
        retrieves the cable delay assuming the ideal resonance has a circular shape
        modifies the cable delay until the shape Im(S21) vs Re(S21) is circular
        see "do_calibration"
        '''
        maxval = np.max(np.absolute(z_data))
        z_data = z_data/maxval
        A1, A2, A3, A4, fr, Ql = self._fit_skewed_lorentzian(f_data,z_data)
        if ignoreslope==True:
            A2 = 0
        else:
            z_data = (np.absolute(z_data)-A2*(f_data-fr)) * np.exp(np.angle(z_data)*1j)  #usually not necessary
        if delay==None:
            if guess==True:
                delay = self._guess_delay(f_data,z_data)
            else:
                delay=0.
            delay = self._fit_delay(f_data,z_data,delay,maxiter=200)
        params = [A1, A2, A3, A4, fr, Ql]
        return delay, params    
    
    def do_calibration(self,f_data,z_data,ignoreslope=True,guessdelay=True):
        '''
        performs an automated calibration and tries to determine the prefactors a, alpha, delay
        fr, Ql, and a possible slope are extra information, which can be used as start parameters for subsequent fits
        see also "do_normalization"
        the calibration procedure works for transmission line resonators as well
        '''
        delay, params = self.get_delay(f_data,z_data,ignoreslope=ignoreslope,guess=guessdelay)
        z_data = (z_data-params[1]*(f_data-params[4]))*np.exp(2.*1j*np.pi*delay*f_data)
        xc, yc, r0 = self._fit_circle(z_data)
        zc = np.complex(xc,yc)
        fitparams = self._phase_fit(f_data,self._center(z_data,zc),0.,np.absolute(params[5]),params[4])
        theta, Ql, fr = fitparams
        beta = self._periodic_boundary(theta+np.pi,np.pi)
        offrespoint = np.complex((xc+r0*np.cos(beta)),(yc+r0*np.sin(beta)))
        alpha = np.angle(offrespoint)
        a = np.absolute(offrespoint)
        return delay, a, alpha, fr, Ql, params[1], params[4]
    
    def do_normalization(self,f_data,z_data,delay,amp_norm,alpha,A2,frcal):
        '''
        removes the prefactors a, alpha, delay and returns the calibrated data, see also "do_calibration"
        works also for transmission line resonators
        '''
        return (z_data-A2*(f_data-frcal))/amp_norm*np.exp(1j*(-alpha+2.*np.pi*delay*f_data))

    def circlefit(self,f_data,z_data,fr=None,Ql=None,refine_results=False,calc_errors=True):
        '''
        performs a circle fit on a frequency vs. complex resonator scattering data set
        Data has to be normalized!!
        INPUT:
        f_data,z_data: input data (frequency, complex S21 data)
        OUTPUT:
        outpus a dictionary {key:value} consisting of the fit values, errors and status information about the fit
        values: {"phi0":phi0, "Ql":Ql, "absolute(Qc)":absQc, "Qi": Qi, "electronic_delay":delay, "complexQc":complQc, "resonance_freq":fr, "prefactor_a":a, "prefactor_alpha":alpha}
        errors: {"phi0_err":phi0_err, "Ql_err":Ql_err, "absolute(Qc)_err":absQc_err, "Qi_err": Qi_err, "electronic_delay_err":delay_err, "resonance_freq_err":fr_err, "prefactor_a_err":a_err, "prefactor_alpha_err":alpha_err}
        for details, see:
            [1] (not diameter corrected) Jiansong Gao, "The Physics of Superconducting Microwave Resonators" (PhD Thesis), Appendix E, California Institute of Technology, (2008)
            [2] (diameter corrected) M. S. Khalil, et. al., J. Appl. Phys. 111, 054510 (2012)
            [3] (fitting techniques) N. CHERNOV AND C. LESORT, "Least Squares Fitting of Circles", Journal of Mathematical Imaging and Vision 23, 239, (2005)
            [4] (further fitting techniques) P. J. Petersan, S. M. Anlage, J. Appl. Phys, 84, 3392 (1998)
        the program fits the circle with the algebraic technique described in [3], the rest of the fitting is done with the scipy.optimize least square fitting toolbox
        also, check out [5] S. Probst et al. "Efficient and reliable analysis of noisy complex scatterung resonator data for superconducting quantum circuits" (in preparation)
        '''
    
        if fr==None: fr=f_data[np.argmin(np.absolute(z_data))]
        if Ql==None: Ql=1e6
        xc, yc, r0 = self._fit_circle(z_data,refine_results=refine_results)
        phi0 = -np.arcsin(yc/r0)
        theta0 = self._periodic_boundary(phi0+np.pi,np.pi)
        z_data_corr = self._center(z_data,np.complex(xc,yc))
        theta0, Ql, fr = self._phase_fit(f_data,z_data_corr,theta0,Ql,fr)
        #print("Ql from phasefit is: " + str(Ql))
        absQc = Ql/(2.*r0)
        complQc = absQc*np.exp(1j*((-1.)*phi0))
        Qc = 1./(1./complQc).real   # here, taking the real part of (1/complQc) from diameter correction method
        Qi_dia_corr = 1./(1./Ql-1./Qc)
        Qi_no_corr = 1./(1./Ql-1./absQc)
    
        results = {"Qi_dia_corr":Qi_dia_corr,"Qi_no_corr":Qi_no_corr,"absQc":absQc,"Qc_dia_corr":Qc,"Ql":Ql,"fr":fr,"theta0":theta0,"phi0":phi0}
    
        #calculation of the error
        p = [fr,absQc,Ql,phi0]
        #chi_square, errors = rt.get_errors(rt.residuals_notch_ideal,f_data,z_data,p)
        if calc_errors==True:
            chi_square, cov = self._get_cov_fast_notch(f_data,z_data,p)
            #chi_square, cov = rt.get_cov(rt.residuals_notch_ideal,f_data,z_data,p)
    
            if cov is not None:
                errors = np.sqrt(np.diagonal(cov))
                fr_err,absQc_err,Ql_err,phi0_err = errors
                #calc Qi with error prop (sum the squares of the variances and covariaces)
                dQl = 1./((1./Ql-1./absQc)**2*Ql**2)
                dabsQc = - 1./((1./Ql-1./absQc)**2*absQc**2)
                Qi_no_corr_err = np.sqrt((dQl**2*cov[2][2]) + (dabsQc**2*cov[1][1])+(2*dQl*dabsQc*cov[2][1]))  #with correlations
                #calc Qi dia corr with error prop
                dQl = 1/((1/Ql-np.cos(phi0)/absQc)**2 *Ql**2)
                dabsQc = -np.cos(phi0)/((1/Ql-np.cos(phi0)/absQc)**2 *absQc**2)
                dphi0 = -np.sin(phi0)/((1/Ql-np.cos(phi0)/absQc)**2 *absQc)
                ##err1 = ( (dQl*cov[2][2])**2 + (dabsQc*cov[1][1])**2 + (dphi0*cov[3][3])**2 )
                err1 = ( (dQl**2*cov[2][2]) + (dabsQc**2*cov[1][1]) + (dphi0**2*cov[3][3]) )
                err2 = ( dQl*dabsQc*cov[2][1] + dQl*dphi0*cov[2][3] + dabsQc*dphi0*cov[1][3] )
                Qi_dia_corr_err =  np.sqrt(err1+2*err2)  # including correlations
                errors = {"phi0_err":phi0_err, "Ql_err":Ql_err, "absQc_err":absQc_err, "fr_err":fr_err,"chi_square":chi_square,"Qi_no_corr_err":Qi_no_corr_err,"Qi_dia_corr_err": Qi_dia_corr_err}
                results.update( errors )
            else:
                print("WARNING: Error calculation failed!")
        else:
            #just calc chisquared:
            fun2 = lambda x: self._residuals_notch_ideal(x,f_data,z_data)**2
            chi_square = 1./float(len(f_data)-len(p)) * (fun2(p)).sum()
            errors = {"chi_square":chi_square}
            results.update(errors)
    
        return results
        
    def autofit(self):
        '''
        automatic calibration and fitting
        '''
        delay, amp_norm, alpha, fr, Ql, A2, frcal =\
                self.do_calibration(self.f_data,self.z_data_raw,ignoreslope=True,guessdelay=True)
        self.z_data = self.do_normalization(self.f_data,self.z_data_raw,delay,amp_norm,alpha,A2,frcal)
        self.fitresults = self.circlefit(self.f_data,self.z_data,fr,Ql,refine_results=False,calc_errors=True)
        self.z_data_sim = A2*(self.f_data-frcal)+self._S21_notch(self.f_data,fr=self.fitresults["fr"],Ql=self.fitresults["Ql"],Qc=self.fitresults["absQc"],phi=self.fitresults["phi0"],a=amp_norm,alpha=alpha,delay=delay)
    
    def _S21_notch(self,f,fr=10e9,Ql=900,Qc=1000.,phi=0.,a=1.,alpha=0.,delay=.0):
        '''
        full model for notch type resonances
        '''
        return a*np.exp(np.complex(0,alpha))*np.exp(-2j*np.pi*f*delay)*(1.-Ql/Qc*np.exp(1j*phi)/(1.+2j*Ql*(f-fr)/fr))    
    
    def get_single_photon_limit(self,unit='dBm'):
        '''
        returns the amout of power in units of W necessary
        to maintain one photon on average in the cavity
        unit can be 'dBm' or 'watt'
        '''
        if self.fitresults!={}:
            fr = self.fitresults['fr']
            k_c = fr/self.fitresults['absQc']
            k_i = fr/self.fitresults['Qi_dia_corr']
            if unit=='dBm':
                return Watt2dBm(1./(4.*k_c/(2.*np.pi*hbar*fr*(k_c+k_i)**2)))
            elif unit=='watt':
                return 1./(4.*k_c/(2.*np.pi*hbar*fr*(k_c+k_i)**2))                
        else:
            warnings.warn('Please perform the fit first',UserWarning)
            return None
        
    def get_photons_in_resonator(self,power,unit='dBm'):
        '''
        returns the average number of photons
        for a given power in units of W
        unit can be 'dBm' or 'watt'
        '''
        if self.fitresults!={}:
            if unit=='dBm':
                power = dBm2Watt(power)
            fr = self.fitresults['fr']
            k_c = fr/self.fitresults['Qc']
            k_i = fr/self.fitresults['Qi']
            return 4.*k_c/(2.*np.pi*hbar*fr*(k_c+k_i)**2) * power
        else:
            warnings.warn('Please perform the fit first',UserWarning)
            return None   

class transmission_port(circlefit,save_load,plotting):
    '''
    a class for handling transmission measurements
    '''
    
    def __init__(self,f_data=None,z_data_raw=None):
        self.porttype = 'transm'
        self.fitresults = {}
        if f_data!=None:
            self.f_data = np.array(f_data)
        else:
            self.f_data=None
        if z_data_raw!=None:
            self.z_data_raw = np.array(z_data_raw)
        else:
            self.z_data=None
        
    def _S21(self,f,fr,Ql,A):
        return A**2/(1.+4.*Ql**2*((f-fr)/fr)**2) 
        
    def fit(self):
        self.ampsqr = (np.absolute(self.z_data_raw))**2
        p = [self.f_data[np.argmax(self.ampsqr)],1000.,np.amax(self.ampsqr)]
        popt, pcov = spopt.curve_fit(self._S21, self.f_data, self.ampsqr,p)
        errors = np.sqrt(np.diag(pcov))
        self.fitresults = {'fr':popt[0],'fr_err':errors[0],'Ql':popt[1],'Ql_err':errors[1],'Ampsqr':popt[2],'Ampsqr_err':errors[2]} 
    
class resonator(object):
    '''
    Universal resonator analysis class
    It can handle different kinds of ports and assymetric resonators.
    '''
    def __init__(self, ports = {}, comment = None):
        '''
        initializes the resonator class object
        ports (dictionary {key:value}): specify the name and properties of the coupling ports
            e.g. ports = {'1':'direct', '2':'notch'}
        comment: add a comment
        '''
        self.comment = comment
        self.port = {}
        self.transm = {}
        if len(ports) > 0:
            for key, pname in ports.iteritems():
                if pname=='direct':
                    self.port.update({key:reflection_port()})
                elif pname=='notch':
                    self.port.update({key:notch_port()})
                else:
                    warnings.warn("Undefined input type! Use 'direct' or 'notch'.", SyntaxWarning)
        if len(self.port) == 0: warnings.warn("Resonator has no coupling ports!", UserWarning)
            
    def add_port(self,key,pname):
        if pname=='direct':
            self.port.update({key:reflection_port()})
        elif pname=='notch':
            self.port.update({key:notch_port()})
        else:
            warnings.warn("Undefined input type! Use 'direct' or 'notch'.", SyntaxWarning)
        if len(self.port) == 0: warnings.warn("Resonator has no coupling ports!", UserWarning)
            
    def delete_port(self,key):
        del self.port[key]
        if len(self.port) == 0: warnings.warn("Resonator has no coupling ports!", UserWarning)
        
    def get_Qi(self):
        '''
        based on the number of ports and the corresponding measurements
        it calculates the internal losses
        '''
        pass
    
    def get_single_photon_limit(self,port):
        '''
        returns the amout of power necessary to maintain one photon 
        on average in the cavity
        '''
        pass
    
    def get_photons_in_resonator(self,power,port):
        '''
        returns the average number of photons
        for a given power
        '''
        pass
        
    def add_transm_meas(self,port1, port2):
        '''
        input: port1
        output: port2
        adds a transmission measurement 
        connecting two direct ports S21
        '''
        key = port1 + " -> " + port2
        self.port.update({key:transm()})
        pass

   
class batch_processing(object):
    '''
    A class for batch processing of resonator data as a function of another variable
    Typical applications are power scans, magnetic field scans etc.
    '''
    
    def __init__(self,porttype):
        '''
        porttype = 'notch', 'direct', 'transm'
        results is an array of dictionaries containing the fitresults
        '''
        self.porttype = porttype
        self.results = []
    
    def autofit(self,cal_dataslice = 0):
        '''
        fits all data
        cal_dataslice: choose scatteringdata which should be used for calibration
        of the amplitude and phase, default = 0 (first)
        '''
        pass
    
class coupled_resonators(batch_processing):
    '''
    A class for fitting a resonator coupled to a second one
    '''
    
    def __init__(self,porttype):
        self.porttype = porttype
        self.results = []
    


# -*- coding: utf-8 -*-
# IV_curve.py analysis class for IV like transport measurement data
# Micha Wildermuth, micha.wildermuth@kit.edu 2019

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import sys
import numpy as np
import matplotlib.pyplot as plt
import scipy as sp
from scipy import stats, signal as sig, optimize as opt, constants as const
const.Phi_0 = const.h / (2 * const.e)

import json
from qkit.measure.json_handler import QkitJSONEncoder, QkitJSONDecoder
from qkit.analysis.qdata import qData, dict2obj

""" Error calculations with uncertainties package """
import uncertainties as uncert
from uncertainties import ufloat, unumpy as unp
from qkit.analysis.uncertainties_utilities import sem, nansem, umean, unanmean, uaverage, unanaverage
np.sem, np.nansem = sem, nansem
uncert.umean, uncert.unanmean = umean, unanmean
uncert.uaverage, uncert.unanaverage = uaverage, unanaverage


class IV_curve3(qData):
    """
    This is an analysis class for IV-like transport measurements taken by `qkit.measure.transport.transport.py`.
    """

    def __init__(self):
        """
        Initializes an analysis class for IV-like transport measurements taken by `qkit.measure.transport.transport.py`.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Examples
        --------
        >>> import numpy as np
        >>> import qkit
        QKIT configuration initialized -> available as qkit.cfg[...]
        >>> qkit.start()
        Starting QKIT framework ... -> qkit.core.startup
        Loading module ... S10_logging.py
        Loading module ... S12_lockfile.py
        Loading module ... S14_setup_directories.py
        Loading module ... S20_check_for_updates.py
        Loading module ... S25_info_service.py
        Loading module ... S30_qkit_start.py
        Loading module ... S65_load_RI_service.py
        Loading module ... S70_load_visa.py
        Loading module ... S80_load_file_service.py
        Loading module ... S85_init_measurement.py
        Loading module ... S98_started.py
        Loading module ... S99_init_user.py
        Initialized the file info database (qkit.fid) in 0.000 seconds.

        >>> from qkit.analysis.IV_curve import IV_curve as IVC
        >>> ivc = IVC()
        Initialized the file info database (qkit.fid) in 0.000 seconds.
        """
        # qkit.fid.update_file_db()  # update file database
        super().__init__()
        self.sweeptype = self.sweeps = self.bias = None
        self.I = self.V = self.V_corr = self.dVdI = self.R_n = self.d2VdI2 = None
        self.I_offsets = self.V_offsets = self.I_offset = self.V_offset = None
        self.scm = self.switching_current(sweeps=self.sweeps,
                                          settings=self.settings)  # subclass for switching current measurement analysis

    def load(self, uuid, dVdI='analysis0'):
        """
        Loads transport measurement data with given uuid <uuid>.

        Parameters
        ----------
        uuid: str
            Qkit identification name, that is looked for and loaded
        dVdI: str | boolean
            Folder, where numerical derivative dV/dI is tried to load form datafile, if this was already analyzed during
            the measurement. If False, dV/dI is not loaded. Default is 'analysis0'.

        Returns
        -------
        None

        Examples
        --------
        >>> ivc.load(uuid='XXXXXX')
        """
        super().load(uuid=uuid)
        if self.m_type != 'transport':
            raise AttributeError('No transport data loaded. Use data acquired with transport measurement class or general qData class.')
        self.scan_dim = self.df.data.i_0.attrs['ds_type']  # scan dimension (1D, 2D, ...)
        self.bias = self.get_bias()
        try:
            self.sweeps = self.measurement.sample.sweeps  # sweeps (start, stop, step)
            self.sweeptype = self.get_sweeptype()
        except AttributeError:
            self.sweeps = [sample.sweeps for sample in self.measurement.sample]
            self.sweeptype = np.unique([self.get_sweeptype(sweeps=sweep) for sweep in self.sweeps])
        shape = np.concatenate([[len(self.sweeps)],
                                np.max([self.df['entry/data0/i_{:d}'.format(j)].shape for j in range(len(self.sweeps))],
                                       axis=0)])  # (number of sweeps, eventually len y-values, eventually len x-values, maximal number of sweep points)
        self.I, self.V, self.dVdI = np.empty(shape=shape), np.empty(shape=shape), np.empty(shape=shape)
        for j in range(shape[0]):
            i = self.df['entry/data0/i_{:d}'.format(j)][:]
            v = self.df['entry/data0/v_{:d}'.format(j)][:]
            if dVdI:
                try:
                    dvdi = self.df['entry/{:s}/dvdi_{:d}'.format(dVdI, j)][:]  # if analysis already done during measurement
                except KeyError:
                    dvdi = self.get_dydx(x=self.I[j], y=self.V[j])
            pad_width = np.insert(np.diff([i.shape, shape[1:]], axis=0),
                                  (0,),
                                  np.zeros(self.scan_dim)).reshape(self.scan_dim, 2)
            if pad_width.any():
                self.I[j] = np.pad(i, pad_width, 'constant', constant_values=np.nan)  # fill with current values (eventually add nans at the end, if sweeps have different lengths)
                self.V[j] = np.pad(v, pad_width, 'constant', constant_values=np.nan)  # fill with voltage values (eventually add nans at the end, if sweeps have different lengths)
                if dVdI:
                    self.dVdI[j] = np.pad(dvdi, pad_width, 'constant', constant_values=np.nan)  # fill with differential resistance values (eventually add nans at the end, if sweeps have different lengths)
            else:
                self.I[j] = i
                self.V[j] = v
                if dVdI:
                    self.dVdI[j] = dvdi
        self._get_xy_parameter(self.df.data.i_0)

        self.scm.sweeps = self.sweeps
        self.scm.settings = self.settings

    def save(self, filename, params=None):
        """
        Saves the class variables
             * uuid,
             * path,
             * measurement_object,
             * measurement_type,
             * scan_dimension,
             * sweep_type,
             * sweeps,
             * bias,
             * I,
             * V,
             * V_corr (only if set
             * dVdI,
             * I_offsets,
             * V_offsets,
             * I_offset,
             * V_offset,
             * x_coordname,
             * x_unit,
             * x_vector,
             * y_coordname,
             * y_unit,
             * y_vector
        as well as <params> to a json-file.

        Parameters
        ----------
        filename: str
            Filename of the .json-file that is created.
        params: dict
            Additional variables that are saved. Default is None, so that only the class variables mentioned above are saved.

        Returns
        -------
        None
        """
        params = params if params else {}
        params = {**{'uuid': self.uuid,
                     'path': self.path,
                     'measurement_object': self.measurement.get_JSON(),
                     'measurement_type': self.m_type,
                     'scan_dimension': self.scan_dim,
                     'sweep_type': self.sweeptype,
                     'sweeps': self.sweeps,
                     'bias': self.bias,
                     'I': self.I,
                     'V': self.V,
                     'dVdI': self.dVdI,
                     'I_offsets': self.I_offsets,
                     'V_offsets': self.V_offsets,
                     'I_offset': self.I_offset,
                     'V_offset': self.V_offset,
                     'x_coordname': self.x_coordname,
                     'x_unit': self.x_unit,
                     'x_vector': self.x_vec,
                     'y_coordname': self.y_coordname,
                     'y_unit': self.y_unit,
                     'y_vector': self.y_vec},
                  **params}
        if self.V_corr:
            params['V_corr'] = self.V_corr
        if '.json' not in filename:
            filename += '.json'
        with open(filename, 'w') as filehandler:
            json.dump(obj=params, fp=filehandler, indent=4, cls=QkitJSONEncoder, sort_keys=True)

    def open_qviewkit(self, uuid=None, ds=None):
        """
        Opens qkit measurement data with given uuid <uuid> in qviewkit.

        Parameters
        ----------
        uuid: str
            Qkit identification name, that is looked for and opened in qviewkit
        ds: str | list(str)
            Datasets that are opened instantaneously. Default is data 'views/IV' and sense values for 2D or 3D.

        Returns
        -------
        None
        """
        if ds is None:
            ds = ['views/IV']
            if self.scan_dim > 1:  # for 2D and 3D: open sense values
                for i in range(len(self.sweeps)):
                    y = '{:s}_{:d}'.format({0: 'I', 1: 'V'}[not self.bias].lower(), i)
                    if y in self.df.data.__dict__.keys():
                        ds.append(y)
        super().open_qviewkit(uuid=uuid, ds=ds)

    def get_bias(self, df=None):
        """
        Gets bias mode of the measurement. Evaluate 'x_ds_url' (1D), 'y_ds_url' (2D), 'z_ds_url' (3D) of i_0 and v_0 and
        checks congruence.

        Parameters
        ----------
        df: qkit.storage.store.Data (optional)
            Datafile of transport measurement. Default is None that means <self.df>.

        Returns
        -------
        mode: int
            Bias mode. Meanings are 0 (current) and 1 (voltage).
        """
        #
        if df is None:
            df = self.df
        self.bias = {'i': 0, 'v': 1}[str(df.data.i_0.attrs.get('{:s}_ds_url'.format(chr(self.scan_dim + 119))) and
                                         df.data.v_0.attrs.get('{:s}_ds_url'.format(chr(self.scan_dim + 119)))).split('/')[-1][0]]
        return self.bias

    def get_sweeptype(self, sweeps=None):
        """
        Gets the sweep type of predefined set of sweeps as generated by `qkit.measure.transport.transport.py`.

        Parameters
        ----------
        sweeps: array_likes of array_likes of floats (optional)
            Set of sweeps containing start, stop and step size (e.g. sweep object using 
            `qkit.measure.transport.transport.sweep` class). Default is None that means self.sweeps.

        Returns
        -------
        sweeptype: int
            Type of set of sweeps. Is 0 (halfswing), 1 (4 quadrants) or None (arbitrary set of sweeps).
        """
        if sweeps is None:
            sweeps = self.sweeps
        # check if sweeps are halfswing
        if len(sweeps) == 2:
            if all(np.array(sweeps[0])[[1, 0, 2]] == np.array(sweeps[1])[:3]):
                self.sweeptype = 0
        # check if sweeps are 4quadrants
        elif len(sweeps) == 4:
            if all(np.array(sweeps[0])[[1, 0, 2]] == np.array(sweeps[1])[:3]) and \
               all(np.array(sweeps[2])[[1, 0, 2]] == np.array(sweeps[3])[:3]):
                self.sweeptype = 1
        else:
            self.sweeptype = None
        return self.sweeptype

    def get_dVdI(self, I=None, V=None, mode=sig.savgol_filter, **kwargs):
        """
        Calculates numerical derivative dV/dI.

        Parameters
        ----------
        I: numpy.array (optional)
            An N-dimensional array containing current values. Default is None that means <self.I>.
        V: numpy.array (optional)
            An N-dimensional array containing voltage values. Default is None that means <self.V>.
        mode: function (optional)
            Function that calculates the numerical gradient dx from a given array x.
            Default is `scipy.signal.savgol_filter` (Savitzky Golay filter).
        kwargs:
            Keyword arguments forwarded to the function <mode>.

        Returns
        -------
        dVdI: numpy.array
            Numerical derivative dV/dI

        Examples
        --------
        # Savitzky Golay filter
        >>> ivc.get_dVdI()

        # numerical gradient
        >>> ivc.get_dVdI(mode=np.gradient)
        """
        if V is None:
            y = self.V
        else:
            y = V
        if I is None:
            x = self.I
        else:
            x = I
        self.dVdI = self.get_dydx(y=y, x=x, mode=mode, **kwargs)
        return self.dVdI

    def get_dydx(self, y, x=None, mode=sig.savgol_filter, **kwargs):
        """
        Calculates numerical derivative dy/dx

        Parameters
        ----------
        y: numpy.array
            An N-dimensional array containing y-values
        x: numpy.array (optional)
            An N-dimensional array containing x-values. Default is None which means that <x> is considered as index.
        mode: function (optional)
            Function that calculates the numerical gradient dx from a given array <x>.
            Default is `scipy.signal.savgol_filter` (Savitzky Golay filter).
        kwargs:
            Keyword arguments forwarded to the function <mode>. Default is for 
            * `scipy.signal.savgol_filter`: {'window_length': 15, 'polyorder': 3, 'deriv': 1},
            * `numpy.gradient`: {'axis': <self.scan_dim>}

        Returns
        -------
        dy/dx: numpy.array
            Numerical gradient quotient. If no <x> is given, dx = np.ones(nop)

        Examples
        --------
        Savitzky Golay filter
        >>> ivc.get_dydx(y=np.arange(1e1), mode=sig.savgol_filter, window_length=9, polyorder=3, deriv=1)
        array([1., 1., 1., 1., 1., 1., 1., 1., 1., 1.])

        numerical gradient
        >>> ivc.get_dydx(mode=np.gradient)
        ivc.get_dydx(y=np.arange(1e1), mode=np.gradient)
        """
        if mode == sig.savgol_filter:
            if 'window_length' not in kwargs.keys():
                kwargs['window_length'] = 15
            if 'polyorder' not in kwargs.keys():
                kwargs['polyorder'] = 3
            if 'deriv' not in kwargs.keys():
                kwargs['deriv'] = 1
        elif mode == np.gradient:
            if 'axis' not in kwargs.keys():
                kwargs['axis'] = self.scan_dim
        if x is None:
            if np.isnan(y).any():
                y_nans = np.isnan(y)
                dy = mode(np.nan_to_num(y, copy=True, nan=0.0), **kwargs)
                np.place(dy, y_nans, np.nan)
                return dy
            else:
                return mode(y, **kwargs)
        else:
            if np.isnan(x).any():
                x_nans = np.isnan(x)  # mask for np.nan
                y_nans = np.isnan(y)
                dx = mode(np.nan_to_num(x, copy=True, nan=0.0), **kwargs)  # derivation function with np.nan replaced by 0
                dy = mode(np.nan_to_num(y, copy=True, nan=0.0), **kwargs)
                np.place(dx, x_nans, np.nan)  # write np.nan using mask from above
                np.place(dy, y_nans, np.nan)
                return dy / dx
            else:
                return mode(y, **kwargs) / mode(x, **kwargs)

    def get_offsets(self, x=None, y=None, dxdy=None, threshold=20e-6, offset=None, yr=False, keepdims=np._NoValue):
        """
        Calculates x- and y-offset for every trace. Therefore, the branch where the y-values are nearly constant are
        evaluated. The average of all corresponding x-values is considered to be the x-offset and the average of the 
        extreme y-values are considered as y-offset.

        Parameters
        ----------
        x: numpy.array (optional)
            An N-dimensional array containing x-values. Default is None, where <x> is considered as <self.V> or <self.I> 
            in the current or voltage bias, respectively.
        y: numpy.array (optional)
            An N-dimensional array containing y-values. Default is None, where <y> is considered as <self.I> or <self.V> 
            in the current or voltage bias, respectively.
        dydx: numpy.array (optional)
            Numerical gradient quotient. Default is None that means
            * the differential resistance <self.dVdI> in the current bias,
            * the differential conductance <self.dIdV> in the voltage bias.
        threshold: float (optional)
            Threshold voltage that limits the superconducting branch. Default is 20e-6.
        offset: float (optional)
            Voltage offset that shifts the limits of the superconducting branch which is set by <threshold>.
            Default is None, which uses voltage at maximal slope dy/dx.
        yr: bool (optional)
            Condition, if critical or retrapping y-values are evaluated. Default is False.
        keepdims : bool (optional)
            If this is set to True, the axes which are reduced are left in the result as dimensions with size one. With
            this option, the result will broadcast correctly against the input array.

            If the default value is passed, then `keepdims` will not be passed through to the `mean` method of
            subclasses of `ndarray`, however any non-default value will be.  If the subclass' method does not implement
            `keepdims` any exceptions will be raised.


        Returns
        -------
        I_offsets: numpy.array
            Current offsets of every single trace.
        V_offsets: numpy.array
            Voltage offsets of every single trace.
        """
        if x is None:
            x = [self.V, self.I][self.bias]
        if y is None:
            y = [self.I, self.V][self.bias]
        if dxdy is None:
            dxdy = [self.dVdI, 1/self.dVdI][self.bias]
        if offset is None:
            amax = np.argmax(1 / sp.ndimage.uniform_filter(dxdy, size=3, mode='constant') if dxdy is not None
                   else np.grandient(y)/np.grandient(x), axis=-1)  # maximal slope
            if len(x.shape) - 1 == 1:
                offset = np.nanmean(x[np.indices(amax.shape)[0], amax])
            elif len(x.shape) - 1 == 2:
                offset = np.nanmean(x[np.indices(amax.shape)[0], np.indices(amax.shape)[1], amax])
            elif len(x.shape) - 1 == 3:
                offset = np.nanmean(x[np.indices(amax.shape)[0], np.indices(amax.shape)[1], np.indices(amax.shape)[2], amax])
        ''' constant range via threshold (for JJ superconducting range via threshold voltage) '''
        with np.errstate(invalid='ignore'):  # raises warning due to np.nan
            mask = np.logical_and(x >= -threshold + offset, x <= threshold + offset)
        x_const, y_const = np.copy(x), np.copy(y)
        np.place(x_const, np.logical_not(mask), np.nan)
        np.place(y_const, np.logical_not(mask), np.nan)
        if self.sweeptype == 0:  # halfswing
            ''' get x offset (for JJ voltage offset)'''
            # x_offsets = np.mean(np.nanmean(x_const, axis=self.scan_dim), axis=0)
            x_offsets = np.atleast_1d(uncert.unanaverage(unp.uarray(nominal_values=np.nanmean(x_const,
                                                                                              axis=self.scan_dim),
                                                                    std_devs=np.nansem(x_const,
                                                                                       axis=self.scan_dim)),
                                                         axis=0))
            ''' get y offset (for JJ current offset) '''
            # TODO: calculate std_devs correct
            if yr:  # retrapping y (for JJ retrapping current)
                y_rs = np.array([np.nanmax(y_const[0], axis=(self.scan_dim - 1)),
                                 np.nanmin(y_const[1], axis=(self.scan_dim - 1))])
                # y_offsets = np.mean(y_rs, axis=0)
                y_offsets = unp.uarray(nominal_values=np.nanmean(y_rs, axis=0),
                                       std_devs=np.nansem(np.abs(y_rs), axis=0))
            else:  # critical y (for JJ critical current)
                y_cs = np.array([np.nanmin(y_const[0], axis=(self.scan_dim - 1)),
                                 np.nanmax(y_const[1], axis=(self.scan_dim - 1))])
                # y_offsets = np.mean(y_cs, axis=0)
                y_offsets = unp.uarray(nominal_values=np.nanmean(y_cs, axis=0),
                                       std_devs=np.nansem(np.abs(y_cs), axis=0))
        elif self.sweeptype == 1:  # 4 quadrants
            ''' get x offset '''
            # x_offsets = np.mean(np.nanmean(x_const, axis=self.scan_dim), axis=0)
            x_offsets = np.atleast_1d(
                uncert.unanaverage(unp.uarray(nominal_values=np.nanmean(x_const,
                                                                        axis=self.scan_dim),
                                              std_devs=np.nansem(x_const,
                                                                 axis=self.scan_dim)),
                                   axis=0))
            ''' get y offset '''
            # TODO: calculate std_devs correct
            if yr:  # retrapping y (for JJ retrapping current)
                y_rs = np.array([np.nanmax(y_const[1], axis=(self.scan_dim - 1)),
                                 np.nanmin(y_const[3], axis=(self.scan_dim - 1))])
                # y_offsets = np.mean(y_rs, axis=0)
                y_offsets = unp.uarray(nominal_values=np.nanmean(y_rs, axis=0),
                                       std_devs=np.nansem(np.abs(y_rs), axis=0))
            else:  # critical y (for JJ critical current)
                y_cs = np.array([np.nanmax(y_const[0], axis=(self.scan_dim - 1)),
                                 np.nanmin(y_const[2], axis=(self.scan_dim - 1))])
                # y_offsets = np.mean(y_cs, axis=0)
                y_offsets = unp.uarray(nominal_values=np.nanmean(y_cs, axis=0),
                                       std_devs=np.nansem(np.abs(y_cs), axis=0))
        else:  # custom sweep type
            ''' get x offset '''
            # x_offsets = np.nanmean(np.nanmean(x_const, axis=self.scan_dim), axis=0)
            x_offsets = np.atleast_1d(
                uncert.unanaverage(unp.uarray(nominal_values=np.nanmean(x_const, axis=self.scan_dim),
                                              std_devs=np.nansem(x_const, axis=self.scan_dim)),
                                   axis=0))
            ''' get y offset '''
            # TODO: calculate std_devs correct
            if yr:  # retrapping y (for JJ retrapping current)
                raise NotImplementedError('No algorithm implemented for custom sweep type')
            else:
                y_cs = np.array([np.nanmax(np.nanmax(y_const, axis=0), axis=self.scan_dim - 1),
                                 np.nanmin(np.nanmin(y_const, axis=0), axis=self.scan_dim - 1)])
                # y_offsets = np.mean(y_cs, axis=0)
                y_offsets = unp.uarray(nominal_values=np.nanmean(y_cs, axis=0),
                                       std_devs=np.nansem(np.abs(y_cs), axis=0))
        self.I_offsets, self.V_offsets = [x_offsets, y_offsets][::int(np.sign(self.bias - .5))]
        if keepdims is not np._NoValue:
            if len(x.shape) - 1 == 1:
                self.I_offsets = self.I_offsets[np.newaxis]
                self.V_offsets = self.V_offsets[np.newaxis]
            elif len(x.shape) - 1 == 2:
                self.I_offsets = self.I_offsets[np.newaxis, :, np.newaxis]
                self.V_offsets = self.V_offsets[np.newaxis, :, np.newaxis]
            elif len(x.shape) - 1 == 3:
                self.I_offsets = self.I_offsets[np.newaxis, :, :, np.newaxis]
                self.V_offsets = self.V_offsets[np.newaxis, :, :, np.newaxis]
        return self.I_offsets, self.V_offsets

    def get_offset(self, *args, **kwargs):
        """
        Calculates x- and y-offset for the whole data set. Therefore, the branch where the y-values are nearly constant
        are evaluated. The average of all corresponding x-values is considered to be the x-offset and the average of the
        extreme y-values are considered as y-offset.

        Parameters
        ----------
        x: numpy.array (optional)
            An N-dimensional array containing x-values. Default is None, where <x> is considered as <self.V> or <self.I>
            in the current or voltage bias, respectively.
        y: numpy.array (optional)
            An N-dimensional array containing y-values. Default is None, where <y> is considered as <self.I> or <self.V>
            in the current or voltage bias, respectively.
        threshold: float (optional)
            Threshold voltage that limits the superconducting branch. Default is 20e-6.
        offset: float (optional)
            Voltage offset that shifts the limits of the superconducting branch which is set by <threshold>. Default is 0.
        yr: bool (optional)
            Condition, if critical or retrapping y-values are evaluated. Default is False.

        Returns
        -------
        I_offsets: numpy.array
            Current offsets of the whole data set.
        V_offsets: numpy.array
            Voltage offsets of the whole data set.
        """
        self.I_offset, self.V_offset = list(map(np.nanmean, self.get_offsets(*args, **kwargs)))
        return self.I_offset, self.V_offset

    def get_2wire_slope_correction(self, I=None, V=None, dVdI=None, peak_finder=sig.find_peaks, **kwargs):
        """
        Gets voltage values corrected by an ohmic slope such as occur in 2wire measurements.
        The two maxima in the differential resistivity <dVdI> are identified as critical and retrapping currents. The
        slope of the superconducting regime in between (which should ideally be infinity) is fitted using
        numpy.linalg.qr algorithm and subtracted from the raw data.

        Parameters
        ----------
        I: numpy.array (optional)
            An N-dimensional array containing current values. Default is None that means <self.I>.
        V: numpy.array (optional)
            An N-dimensional array containing voltage values. Default is None that means <self.V>.
        dVdI: numpy.array (optional)
            An N-dimensional array containing differential resistance (dV/dI) values. Default is None that means
            <self.dVdI>.
        peak_finder: function (optional)
            Peak finding algorithm. Default is `scipy.signal.find_peaks`.
        kwargs:
            Keyword arguments forwarded to the peak finding algorithm <peak_finder>.

        Returns
        -------
        V_corr: numpy.array
            Ohmic slope corrected voltage values
        """

        # TODO: uncertainties
        def lin_fit(x, y):
            X = np.stack((x, np.ones(len(x)))).T
            q, r = np.linalg.qr(X)
            p = np.dot(q.T, y)
            return np.dot(np.linalg.inv(r), p)

        if I is None:
            I = self.I
        if V is None:
            V = self.V
        if dVdI is None:
            dVdI = self.dVdI
        if peak_finder is sig.find_peaks and 'prominence' not in kwargs.keys():
            kwargs['prominence'] = 100
        if peak_finder is sig.find_peaks_cwt and 'widths' not in kwargs.keys():
            kwargs['widths'] = np.arange(10)
        ''' peak detection in dV/dI '''
        if len(V.shape) - 1 == 1:
            peaks = map(lambda dVdI1D:
                        peak_finder(dVdI1D, **kwargs),
                        dVdI)
            slices = map(lambda peaks1D:
                         slice(*np.sort(peaks1D[0][peaks1D[1]['prominences'].argsort()[-2:][::-1]])),
                         peaks)
            popts = map(lambda I1D, V1D, s1D:
                        lin_fit(I1D[s1D], V1D[s1D]),
                        I, V, slices)
            self.V_corr = np.array(list(map(lambda I1D, V1D, popt1D:
                                            V1D - (popt1D[0] * I1D + popt1D[1]),
                                            I, V, popts)),
                                   dtype=float)
            return self.V_corr
        elif len(V.shape) - 1 == 2:
            peaks = map(lambda dVdI2D:
                        map(lambda dVdI1D:
                            peak_finder(dVdI1D, **kwargs),
                            dVdI2D),
                        dVdI)
            slices = map(lambda peaks2D:
                         map(lambda peaks1D:
                             slice(*np.sort(peaks1D[0][peaks1D[1]['prominences'].argsort()[-2:][::-1]])),
                             peaks2D),
                         peaks)
            popts = map(lambda I2D, V2D, s2D:
                        map(lambda I1D, V1D, s1D:
                            lin_fit(I1D[s1D], V1D[s1D]),
                            I2D, V2D, s2D),
                        I, V, slices)
            self.V_corr = np.array(list(map(lambda I2D, V2D, popt2D:
                                            map(lambda I1D, V1D, popt1D:
                                                V1D - (popt1D[0] * I1D + popt1D[1]),
                                                I2D, V2D, popt2D),
                                            I, V, popts)),
                                   dtype=float)
            return self.V_corr
        elif len(V.shape) - 1 == 3:
            peaks = map(lambda dVdI3D:
                        map(lambda dVdI2D:
                            map(lambda dVdI1D:
                                peak_finder(dVdI1D, **kwargs),
                                dVdI2D),
                            dVdI3D),
                        dVdI)
            slices = map(lambda peaks3D:
                         map(lambda peaks2D:
                             map(lambda peaks1D:
                                 slice(*np.sort(peaks1D[0][peaks1D[1]['prominences'].argsort()[-2:][::-1]])),
                                 peaks2D),
                             peaks3D),
                         peaks)
            popts = map(lambda I3D, V3D, s3D:
                        map(lambda I2D, V2D, s2D:
                            map(lambda I1D, V1D, s1D:
                                lin_fit(I1D[s1D], V1D[s1D]),
                                I2D, V2D, s2D),
                            I3D, V3D, s3D),
                        I, V, slices)
            self.V_corr = np.array(list(map(lambda I3D, V3D, popt3D:
                                            map(lambda I2D, V2D, popt2D:
                                                map(lambda I1D, V1D, popt1D:
                                                    V1D - (popt1D[0] * I1D + popt1D[1]),
                                                    I2D, V2D, popt2D),
                                                I3D, V3D, popt3D),
                                            I, V, popts)),
                                   dtype=float)
            return self.V_corr
        else:
            raise ValueError('Scan dimension must be in {1, 2, 3}')

    def get_Rn(self, I=None, V=None, dVdI=None, deriv_func=sig.savgol_filter, peak_finder=sig.find_peaks, mode=0,
               **kwargs):
        """
        Get normal state resistance of over critical range. Therefore, the curvature d²V/dI² is computed using the second
        order derivation function <deriv_func> and analysing peaks in it with <peak_finder>.
        The ohmic range is considered to range from the outermost tail of the peaks in the curvature to the start/end of
        the sweep and the resistance is calculated as mean of the differential resistance values <dVdI> within this range.

        Parameters
        ----------
        I: numpy.array (optional)
            An N-dimensional array containing current values. Default is None that means <self.I>.
        V: numpy.array (optional)
            An N-dimensional array containing voltage values. Default is None that means <self.V>.
        dVdI: numpy.array (optional)
            An N-dimensional array containing differential resistance (dV/dI) values. Default is None that means
            <self.dVdI>.
        deriv_func: function (optional)
            Function that calculates the numerical gradient dx from a given array x. Default is
            `scipy.signal.savgol_filter` (Savitzky Golay filter).
        peak_finder: function (optional)
            Peak finding algorithm. Default is `scipy.signal.find_peaks`.
        mode: int (optional)
            Mode how Rn is determined from the ohmic range found. Must be 0 (slope of linear fit) or 1 (average of dV/dI).
        kwargs:
            Keyword arguments forwarded to the function <mode> and the peak finding algorithm <peak_finder>. Default is for
            * `scipy.signal.savgol_filter`: {'window_length': 15, 'polyorder': 3, 'deriv': 2},
            * `numpy.diff`: {'n': 2, 'axis': <self.scan_dim}>,
            * `scipy.signal.find_peaks`: {'prominence': np.max(np.abs(<self.d2VdI2>))/1e2)}

        Returns
        -------
        Rn: numpy.array
            Average normal state resistance
        """

        def _peak_finder(a, **_kwargs):
            ans = peak_finder(a, **_kwargs)
            if np.array_equal(ans[0], []):  # no peaks found
                return [np.array([False]), {}]
            else:
                return ans

        if I is None:
            I = self.I
        if V is None:
            V = self.V
        if dVdI is None:
            dVdI = self.dVdI
        if deriv_func == sig.savgol_filter:
            kwargs_deriv = {'deriv': 2, 'window_length': kwargs.get('window_length', 15),
                            'polyorder': kwargs.get('polyorder', 3)}
        elif deriv_func == np.diff:
            kwargs_deriv = {'n': 2, 'axis': kwargs.get('axis', self.scan_dim)}
        else:
            kwargs_deriv = {}
        ''' second derivative d^2V/dI^2 '''
        self.d2VdI2 = self.get_dydx(x=I, y=V, mode=deriv_func, **kwargs_deriv)
        ''' peak detection in d^2V/dI^2 '''
        if len(V.shape) - 1 == 1:
            if peak_finder == sig.find_peaks:
                kwargs_peak_finder = np.array(list(map(lambda d2VdI21D:
                                                       {'prominence': kwargs.get('prominence',
                                                                                 np.max(np.abs(d2VdI21D)) / 1e2)},
                                                       self.d2VdI2)))
            peaks = map(lambda j, d2VdI21D:
                        _peak_finder(d2VdI21D * np.sign(max(d2VdI21D.min(), d2VdI21D.max(), key=abs)),
                                     **kwargs_peak_finder[j]),  # sign, so that extremum is positiv
                        *zip(*enumerate(self.d2VdI2)))
            slcs = np.array(list(map(lambda peak1D, d2VdI21D:
                                     np.array((slice(0, np.min(peak1D[1]['left_bases'])),
                                               slice(max(peak1D[1]['right_bases']), d2VdI21D.size))),
                                     peaks, self.d2VdI2)))
            if mode == 0:
                popts, pcovs = np.ones(shape=(len(self.d2VdI2), 2)), np.ones(shape=(len(self.d2VdI2), 2, 2))
                for j in range(2):
                    V_ohm = np.concatenate([V[k, slc] for k, slc in enumerate(slcs.T[j])])
                    I_ohm = np.concatenate([I[k, slc] for k, slc in enumerate(slcs.T[j])])
                    popts[j], pcovs[j] = np.polyfit(x=I_ohm,
                                                    y=V_ohm,
                                                    deg=1,
                                                    cov=True)
                # self.R_n = np.mean(popts[:, 0])
                self.R_n = np.nanmean(unp.uarray(nominal_values=popts,
                                                 std_devs=np.array(list(map(lambda pcov:
                                                                            np.sqrt(np.diagonal(pcov)),
                                                                            pcovs))))[:, 0])
            elif mode == 1:
                # self.R_n = np.array([np.nanmean(np.concatenate([dVdI[k, x, y, s] for k, s in enumerate(slc)]))
                #                     for j, slc in enumerate(np.transpose(slcs, axes=(1, 0)))])
                self.R_n = np.array(
                    [ufloat(nominal_value=np.nanmean(np.concatenate([dVdI[k, s] for k, s in enumerate(slc)])),
                            std_dev=np.nansem(np.concatenate([dVdI[k, s] for k, s in enumerate(slc)])))
                     for j, slc in enumerate(np.transpose(slcs, axes=(1, 0)))])
                self.R_n = np.nanmean(self.R_n, axis=self.scan_dim - 1)
        elif len(V.shape) - 1 == 2:
            if peak_finder == sig.find_peaks:
                kwargs_peak_finder = np.array(list(map(lambda d2VdI22D:
                                                       list(map(lambda d2VdI21D:
                                                                {'prominence': kwargs.get('prominence', np.max(
                                                                    np.abs(d2VdI21D)) / 1e2)},
                                                                d2VdI22D)),
                                                       self.d2VdI2)))
            peaks = map(lambda j, d2VdI22D:
                        list(map(lambda k, d2VdI21D:
                                 _peak_finder(d2VdI21D * np.sign(max(d2VdI21D.min(), d2VdI21D.max(), key=abs)),
                                              **kwargs_peak_finder[j, k]),  # sign, so that extremum is positiv
                                 *zip(*enumerate(d2VdI22D)))),
                        *zip(*enumerate(self.d2VdI2)))
            slcs = np.array(list(map(lambda peak2D, d2VdI22D:
                                     list(map(lambda peak1D, d2VdI21D:
                                              np.array((slice(0, np.min(peak1D[1]['left_bases'])),
                                                        slice(max(peak1D[1]['right_bases']), d2VdI21D.size))),
                                              peak2D, d2VdI22D)),
                                     peaks, self.d2VdI2)))
            if mode == 0:
                popts = np.ones(shape=(len(self.x_vec), len(self.d2VdI2), 2))
                pcovs = np.ones(shape=(len(self.x_vec), len(self.d2VdI2), 2, 2))
                for x, slcs1D in enumerate(np.transpose(slcs, axes=(1, 0, 2))):
                    for j in range(2):
                        V_ohm = np.concatenate([V[k, x, slc] for k, slc in enumerate(slcs1D.T[j])])
                        I_ohm = np.concatenate([I[k, x, slc] for k, slc in enumerate(slcs1D.T[j])])
                        popts[x, j], pcovs[x, j] = np.polyfit(x=I_ohm,
                                                              y=V_ohm,
                                                              deg=1,
                                                              cov=True)
                # self.R_n = np.nanmean(popts[:, :, 0], axis=1)
                self.R_n = np.nanmean(unp.uarray(nominal_values=popts,
                                                 std_devs=np.array(list(map(lambda pcov1D:
                                                                            list(map(lambda pcov:
                                                                                     np.sqrt(np.diagonal(pcov)),
                                                                                     pcov1D)),
                                                                            pcovs))))[:, :, 0],
                                      axis=1)
            elif mode == 1:
                # self.R_n = np.array([[np.nanmean(np.concatenate([dVdI[k, x, y, s] for k, s in enumerate(slc)]))
                #                      for j, slc in enumerate(slcs1D)]
                #                     for x, slcs1D in enumerate(np.transpose(slcs, axes=(1, 2, 0)))])
                self.R_n = np.array(
                    [[ufloat(nominal_value=np.nanmean(np.concatenate([dVdI[k, s] for k, s in enumerate(slc)])),
                             std_dev=np.nansem(np.concatenate([dVdI[k, s] for k, s in enumerate(slc)])))
                      for j, slc in enumerate(slcs1D)]
                     for x, slcs1D in enumerate(np.transpose(slcs, axes=(1, 2, 0)))])
                self.R_n = np.nanmean(self.R_n, axis=self.scan_dim - 1)
        elif len(V.shape) - 1 == 3:
            if peak_finder == sig.find_peaks:
                kwargs_peak_finder = np.array(list(map(lambda d2VdI23D:
                                                       list(map(lambda d2VdI22D:
                                                                list(map(lambda d2VdI21D:
                                                                         {'prominence': kwargs.get('prominence', np.max(
                                                                             np.abs(d2VdI21D)) / 1e2)},
                                                                         d2VdI22D)),
                                                                d2VdI23D)),
                                                       self.d2VdI2)))
            peaks = map(lambda i, d2VdI23D:
                        list(map(lambda j, d2VdI22D:
                                 list(map(lambda k, d2VdI21D:
                                          _peak_finder(d2VdI21D * np.sign(max(d2VdI21D.min(), d2VdI21D.max(), key=abs)),
                                                       **kwargs_peak_finder[i, j, k]),
                                          # sign, so that extremum is positiv
                                          *zip(*enumerate(d2VdI22D)))),
                                 *zip(*enumerate(d2VdI23D)))),
                        *zip(*enumerate(self.d2VdI2)))
            slcs = np.array(list(map(lambda peak3D, d2VdI23D:
                                     list(map(lambda peak2D, d2VdI22D:
                                              list(map(lambda peak1D, d2VdI21D:
                                                       np.array((slice(0, np.min(peak1D[1]['left_bases'])),
                                                                 slice(max(peak1D[1]['right_bases']), d2VdI21D.size)
                                                                 )),
                                                       peak2D, d2VdI22D)),
                                              peak3D, d2VdI23D)),
                                     peaks, self.d2VdI2)))
            if mode == 0:
                popts = np.ones(shape=(len(self.x_vec), len(self.y_vec), len(self.d2VdI2), 2))
                pcovs = np.ones(shape=(len(self.x_vec), len(self.y_vec), len(self.d2VdI2), 2, 2))
                for x, slcs2D in enumerate(np.transpose(slcs, axes=(1, 2, 0, 3))):
                    for y, slcs1D in enumerate(slcs2D):
                        for j in range(2):
                            V_ohm = np.concatenate([V[k, x, y, slc] for k, slc in enumerate(slcs1D.T[j])])
                            I_ohm = np.concatenate([I[k, x, y, slc] for k, slc in enumerate(slcs1D.T[j])])
                            popts[x, y, j], pcovs[x, y, j] = np.polyfit(x=I_ohm,
                                                                        y=V_ohm,
                                                                        deg=1,
                                                                        cov=True)
                # self.R_n = np.nanmean(popts[:, :, :, 0], axis=2)
                self.R_n = np.nanmean(unp.uarray(nominal_values=popts,
                                                 std_devs=np.array(list(map(lambda pcov2D:
                                                                            list(map(lambda pcov1D:
                                                                                     list(map(lambda pcov:
                                                                                              np.sqrt(
                                                                                                  np.diagonal(pcov)),
                                                                                              pcov1D)),
                                                                                     pcov2D)),
                                                                            pcovs))))[:, :, :, 0],
                                      axis=2)
            elif mode == 1:
                # self.R_n = np.array([[[np.nanmean(np.concatenate([dVdI[k, x, y, s] for k, s in enumerate(slc)]))
                #                       for j, slc in enumerate(slcs1D)]
                #                      for y, slcs1D in enumerate(slcs2D)]
                #                     for x, slcs2D in enumerate(np.transpose(slcs, axes=(1, 2, 3, 0)))])
                self.R_n = np.array([[[ufloat(nominal_value=np.nanmean(np.concatenate([dVdI[k, s]
                                                                                       for k, s in enumerate(slc)])),
                                              std_dev=np.nansem(np.concatenate([dVdI[k, s]
                                                                                for k, s in enumerate(slc)])))
                                       for j, slc in enumerate(slcs1D)]
                                      for y, slcs1D in enumerate(slcs2D)]
                                     for x, slcs2D in enumerate(np.transpose(slcs, axes=(1, 2, 3, 0)))])
                self.R_n = np.nanmean(self.R_n, axis=self.scan_dim - 1)
        else:
            raise ValueError('Scan dimension must be in {1, 2, 3}')
        return self.R_n

    def get_Ic_threshold(self, I=None, V=None, dVdI=None, threshold=20e-6, offset=None, Ir=False):
        """
        Get critical current values. These are considered as currents, where the voltage jumps beyond threshold
        ± <threshold> - <offset>.

        Parameters
        ----------
        I: numpy.array (optional)
            An N-dimensional array containing current values. Default is None that means <self.I>.
        V: numpy.array (optional)
            An N-dimensional array containing voltage values. Default is None that means <self.V>.
        dVdI: numpy.array (optional)
            An N-dimensional array containing differential resistance (dV/dI) values. Default is None that means
            <self.dVdI>.
        threshold: float (optional)
            Threshold voltage that limits the superconducting branch. Default is 20e-6.
        offset: float (optional)
            Voltage offset that shifts the limits of the superconducting branch which is set by <threshold>. Default is
            None that means <self.V_offset>.
        Ir: bool (optional)
            Condition, if retrapping currents are returned, too. Default is False.

        Returns
        -------
        I_cs: numpy.array
            Critical current values.
        I_rs: numpy.array (optional)
            Retrapping current values.
        """
        # TODO: uncertainties
        if V is None:
            V = self.V
        if I is None:
            I = self.I
        if dVdI is None:
            dVdI = self.dVdI
        if offset is None:
            if self.V_offset is None:
                amax = np.argmax(1 / dVdI if dVdI is not None else np.grandient(I)/np.grandient(V), axis=-1)  # maximal slope
                if len(V.shape) - 1 == 1:
                    offset = np.nanmean(V[np.indices(amax.shape)[0], amax])
                elif len(V.shape) - 1 == 2:
                    offset = np.nanmean(V[np.indices(amax.shape)[0], np.indices(amax.shape)[1], amax])
                elif len(V.shape) - 1 == 3:
                    offset = np.nanmean(V[np.indices(amax.shape)[0], np.indices(amax.shape)[1], np.indices(amax.shape)[2], amax])
            else:
                offset = self.V_offset
        if len(V.shape) - 1 == 0:  # single trace used for in situ fit
            with np.errstate(invalid='ignore'):  # raises warning due to np.nan
                mask = np.logical_and(V >= -threshold + offset, V <= threshold + offset)
            I_sc = np.copy(I)
            np.place(I_sc, np.logical_not(mask), np.nan)
            return np.nanmax(I_sc)
        ''' constant range via threshold (for JJ superconducting range via threshold voltage) '''
        with np.errstate(invalid='ignore'):  # raises warning due to np.nan
            mask = np.logical_and(V >= -threshold + offset, V <= threshold + offset)
        V_sc, I_sc = np.copy(V), np.copy(I)
        np.place(V_sc, np.logical_not(mask), np.nan)
        np.place(I_sc, np.logical_not(mask), np.nan)
        if self.sweeptype == 0:  # halfswing
            ''' critical current '''
            I_cs = np.array([np.nanmin(I_sc[0], axis=(self.scan_dim - 1)),
                             np.nanmax(I_sc[1], axis=(self.scan_dim - 1))])
            ''' retrapping current '''
            I_rs = np.array([np.nanmax(I_sc[0], axis=(self.scan_dim - 1)),
                             np.nanmin(I_sc[1], axis=(self.scan_dim - 1))])
        elif self.sweeptype == 1:  # 4 quadrants
            ''' critical current '''
            I_cs = np.array([np.nanmax(I_sc[0], axis=(self.scan_dim - 1)),
                             np.nanmin(I_sc[2], axis=(self.scan_dim - 1))])
            ''' retrapping current '''
            I_rs = np.array([np.nanmax(I_sc[1], axis=(self.scan_dim - 1)),
                             np.nanmin(I_sc[3], axis=(self.scan_dim - 1))])
        else:  # custom sweep type
            ''' critical current '''
            I_cs = np.array([np.nanmax(I_sc, axis=(self.scan_dim - 1)),
                             np.nanmin(I_sc, axis=(self.scan_dim - 1))])
            ''' retrapping current '''
            if Ir:
                raise NotImplementedError('No algorithm implemented for custom sweep type')
        if Ir:
            return [I_cs, I_rs]
        else:
            return I_cs

    def get_Ic_deriv(self, I=None, V=None, dVdI=None, Ir=False, Vg=False, tol_offset=20e-6, window=5,
                     peak_finder=sig.find_peaks, **kwargs):
        """
        Gets critical current values using the numerical derivative dV/dI.
        Peaks in these data correspond to voltage jumps, are detected with a peak finding algorithm <peak_finder> and
        checked, whether the corresponding voltage jumps out or in the superconducting branch, that is identified as
        critical or retrapping current, respectively. Therefore, the average of half the window below and above the peak
        is considered. The superconducting branch, in turn, is assumed as the voltage offset <self.V_offset> within the
        tolerance <tol_offset>.

        Parameters
        ----------
        I: numpy.array (optional)
            An N-dimensional array containing current values. Default is None that means <self.I>.
        V: numpy.array (optional)
            An N-dimensional array containing voltage values. Default is None that means <self.V>.
        dVdI: numpy.array (optional)
            An N-dimensional array containing differential resistance (dV/dI) values. Default is None that means
            <self.dVdI>.
        Ir: bool (optional)
            Condition, if retrapping currents are returned, too. Default is False
        Vg: bool (optional)
            Condition, if gap voltages are returned, too. Default is False
        tol_offset: float (optional)
            Voltage offset tolerance that limits the superconducting branch around the voltage offset <self.V_offset>.
            Default is 20e-6.
        window: int (optional)
            Window around the jump, where the voltage is evaluated and classified as 'superconducting branch'. Default
            is 5 that considers two values below and 2 values above the jump.
        peak_finder: function (optional)
            Peak finding algorithm. Default is `scipy.signal.find_peaks`.
        kwargs:
            Keyword arguments forwarded to the peak finding algorithm <peak_finder>.

        Returns
        -------
        I_cs: numpy.array
            Critical current values.
        I_rs: numpy.array (optional)
            Retrapping current values.
        properties: dict
            Properties of all found peaks (not only I_c and I_r, but also further jumps), such as corresponding currents,
            voltages, differential resistances, indices as well as returns of the used peak finding algorithm.

        Examples
        --------
        >>> I_cs, props = ivc.get_Ic_deriv(prominence=100)
        >>> if ivc.scan_dim == 1:
        >>>     Is = np.array(list(map(lambda p1D: p1D['I'], props)))  # has shape (number of sweeps, number of peaks)
        >>> elif ivc.scan_dim == 2:
        >>>     Is = np.array(list(map(lambda p2D: list(map(lambda p1D: p1D['I'], p2D)), props)))  # has shape (number of sweeps, number of x-values, number of peaks)
        >>> elif ivc.scan_dim == 3:
        >>>     Is = np.array(list(map(lambda p3D: list(map(lambda p2D: list(map(lambda p1D: p1D['I'], p2D)), p3D)), props)))  # has shape (number of sweeps, number of y-values, number of x-values, number of peaks)
        """

        def _peak_finder(x, **_kwargs):
            ans = peak_finder(x, **_kwargs)
            if np.array_equal(ans[0], []):  # no peaks found
                return [np.array([False]), {}]
            else:
                return ans

        if I is None:
            I = self.I
        if V is None:
            V = self.V
        if dVdI is None:
            dVdI = self.dVdI
        if peak_finder is sig.find_peaks and 'prominence' not in kwargs.keys():
            kwargs['prominence'] = 100
        if peak_finder is sig.find_peaks_cwt and 'widths' not in kwargs.keys():
            kwargs['widths'] = np.arange(10)
        if len(V.shape) - 1 == 0:  # single trace used for in situ fit
            peaks = _peak_finder(dVdI, **kwargs)
            try:
                return I[peaks[0]]
            except IndexError:  # if no peak found return np.nan
                return np.nan
        ''' peak detection in dV/dI '''
        if len(V.shape) - 1 == 1:
            peaks = np.array(list(map(lambda dVdI1D:
                                      _peak_finder(dVdI1D, **kwargs),
                                      dVdI)),
                             dtype=object)
        elif len(V.shape) - 1 == 2:
            peaks = np.array(list(map(lambda dVdI2D:
                                      list(map(lambda dVdI1D:
                                               _peak_finder(dVdI1D, **kwargs),
                                               dVdI2D)),
                                      dVdI)),
                             dtype=object)
        elif len(V.shape) - 1 == 3:
            peaks = np.array(list(map(lambda dVdI3D:
                                      list(map(lambda dVdI2D:
                                               list(map(lambda dVdI1D:
                                                        _peak_finder(dVdI1D, **kwargs),
                                                        dVdI2D)),
                                               dVdI3D)),
                                      dVdI)),
                             dtype=object)
        else:
            raise ValueError('Scan dimension must be in {1, 2, 3}')
        return self._classify_jump(I=I, V=V, Y=dVdI, peaks=peaks, tol_offset=tol_offset, window=window, Ir=Ir, Vg=Vg)

    def get_Ic_dft(self, I=None, V=None, dVdI=None, s=10, Ir=False, Vg=False, tol_offset=20e-6, window=5,
                   peak_finder=sig.find_peaks, **kwargs):
        """
        Gets critical current values using a discrete Fourier transform, a smoothed derivation in the frequency domain
        and an inverse Fourier transform.
        Therefore, the voltage values are corrected by the linear offset slope, fast Fourier transformed to the frequency
        domain, multiplied with a Gaussian smoothed derivation function if*exp(-s*f²) in the frequency domain and
        inversely fast Fourier transformed to the time domain. This corresponds to the convolution of the voltage values
        with the Gaussian smoothed derivation function in the time domain.
        Peaks in these data correspond to voltage jumps, are detected with a peak finding algorithm <peak_finder> and
        checked, whether the corresponding voltage jumps out or in the superconducting branch, that is identified as
        critical or retrapping current, respectively. Therefore, the average of half the window below and above the peak
        is considered. The superconducting branch, in turn, is assumed as the voltage offset <self.V_offset> within the
        tolerance <tol_offset>.

        Parameters
        ----------
        I: numpy.array (optional)
            An N-dimensional array containing current values. Default is None that means <self.I>.
        V: numpy.array (optional)
            An N-dimensional array containing voltage values. Default is None that means <self.V>.
        dVdI: numpy.array (optional)
            An N-dimensional array containing differential resistance (dV/dI) values. Default is None that means
            <self.dVdI>.
        s: float (optional)
            Smoothing factor of the derivative. Default is 10.
        Ir: bool (optional)
            Condition, if retrapping currents are returned, too. Default is False.
        Vg: bool (optional)
            Condition, if gap voltages are returned, too. Default is False.
        tol_offset: float (optional)
            Voltage offset tolerance that limits the superconducting branch around the voltage offset <self.V_offset>.
            Default is 20e-6.
        window: int (optional)
            Window around the jump, where the voltage is evaluated and classified as 'superconducting branch'. Default
            is 5 that considers two values below and 2 values above the jump.
        peak_finder: function (optional)
            Peak finding algorithm. Default is `scipy.signal.find_peaks`.
        kwargs:
            Keyword arguments forwarded to the peak finding algorithm <peak_finder>.

        Returns
        -------
        I_cs: numpy.array
            Critical current values.
        I_rs: numpy.array (optional)
            Retrapping current values.
        properties: dict
            Properties of all found peaks (not only I_c and I_r, but also further jumps), such as corresponding currents,
            voltages, differential resistances, indices as well as returns of the used peak finding algorithm.

        Examples
        --------
        >>> I_cs, props = ivc.get_Ic_dft(prominence=1)
        >>> if ivc.scan_dim == 1:
        >>>     Is = np.array(list(map(lambda p1D: p1D['I'], props)))  # has shape (number of sweeps, number of peaks)
        >>> elif ivc.scan_dim == 2:
        >>>     Is = np.array(list(map(lambda p2D: list(map(lambda p1D: p1D['I'], p2D)), props)))  # has shape (number of sweeps, number of x-values, number of peaks)
        >>> elif ivc.scan_dim == 3:
        >>>     Is = np.array(list(map(lambda p3D: list(map(lambda p2D: list(map(lambda p1D: p1D['I'], p2D)), p3D)), props)))  # has shape (number of sweeps, number of y-values, number of x-values, number of peaks)
        """

        def _get_deriv_dft(_V):
            V_fft = np.fft.fft(_V)  # Fourier transform of V from time to frequency domain
            f = np.fft.fftfreq(_V.shape[-1])  # frequency values
            kernel = 1j * np.fft.fft(f * np.exp(-s * f ** 2))  # smoothed derivation function, how it would look like in time domain (with which V is convolved in the time domain)
            V_fft_smooth = 1j * f * np.exp(-s * f ** 2) * V_fft  # Fourier transform of a Gaussian smoothed derivation of V in the frequency domain
            return np.fft.ifft(V_fft_smooth)  # inverse Fourier transform of the smoothed derivation of V from reciprocal to time domain

        def _peak_finder(x, **_kwargs):
            ans = peak_finder(x, **_kwargs)
            if np.array_equal(ans[0], []):  # no peaks found
                return [np.array([False]), {}]
            else:
                return ans

        if I is None:
            I = self.I
        if V is None:
            V = self.V
        if peak_finder is sig.find_peaks and 'prominence' not in kwargs.keys():
            kwargs['prominence'] = 1e-5
        if peak_finder is sig.find_peaks_cwt and 'widths' not in kwargs.keys():
            kwargs['widths'] = np.arange(10)
        if len(V.shape) - 1 == 0:  # single trace used for in situ fit
            V_corr = V - np.linspace(start=V[0], stop=V[-1], num=V.shape[-1], axis=0)  # adjust offset slope
            dV_smooth = _get_deriv_dft(_V=V_corr)
            peaks = _peak_finder(dV_smooth, **kwargs)
            try:
                return I[peaks[0]]
            except IndexError:  # if no peak found return np.nan
                return np.nan
        ''' differentiate and smooth in the frequency domain '''
        if len(V.shape) - 1 == 1:
            V_corr = V - np.linspace(start=V[:, 0],
                                     stop=V[:, -1],
                                     num=V.shape[-1],
                                     axis=1)  # adjust offset slope
        elif len(V.shape) - 1 == 2:
            V_corr = V - np.linspace(start=V[:, :, 0],
                                     stop=V[:, :, -1],
                                     num=V.shape[-1],
                                     axis=2)  # adjust offset slope
        elif len(V.shape) - 1 == 3:
            V_corr = V - np.linspace(start=V[:, :, :, 0],
                                     stop=V[:, :, :, -1],
                                     num=V.shape[-1],
                                     axis=3)  # adjust offset slope
        else:
            raise ValueError('Scan dimension must be in {1, 2, 3}')
        dV_smooth = _get_deriv_dft(_V=V_corr)
        ''' peak detection '''
        if len(V.shape) - 1 == 1:
            peaks = np.array(list(map(lambda dV_smooth1D:
                                      _peak_finder(np.abs(dV_smooth1D), **kwargs),
                                      dV_smooth)),
                             dtype=object)
        elif len(V.shape) - 1 == 2:
            peaks = np.array(list(map(lambda dV_smooth2D:
                                      list(map(lambda dV_smooth1D:
                                               _peak_finder(np.abs(dV_smooth1D), **kwargs),
                                               dV_smooth2D)),
                                      dV_smooth)),
                             dtype=object)
        elif len(V.shape) - 1 == 3:
            peaks = np.array(list(map(lambda dV_smooth3D:
                                      list(map(lambda dV_smooth2D:
                                               list(map(lambda dV_smooth1D:
                                                        _peak_finder(np.abs(dV_smooth1D), **kwargs),
                                                        dV_smooth2D)),
                                               dV_smooth3D)),
                                      dV_smooth)),
                             dtype=object)
        else:
            raise ValueError('Scan dimension must be in {1, 2, 3}')
        return self._classify_jump(I=I, V=V, Y=dV_smooth, peaks=peaks, tol_offset=tol_offset, window=window, Ir=Ir,
                                   Vg=Vg)

    def _classify_jump(self, I, V, Y, peaks, tol_offset=20e-6, window=5, Ir=False, Vg=False):
        """
        Classifies voltage jumps as critical currents, retrapping currents of none of those.

        Parameters
        ----------
        I: numpy.array
            An N-dimensional array containing current values.
        V: numpy.array
            An N-dimensional array containing voltage values.
        Y: numpy.array
            An N-dimensional array containing data, whose peaks are already determined.
        peaks numpy.array
            An N-dimensional array containing indices and properties of peaks that are already determined, as obtained
            by e.g. `scipy.signal.find_peaks`.
        Ir: bool (optional)
            Condition, if retrapping currents are returned, too. Default is False
        Vg: bool (optional)
            Condition, if gap voltages are returned, too. Default is False
        tol_offset: float (optional)
            Voltage offset tolerance that limits the superconducting branch around the voltage offset <self.V_offset>.
            Default is 20e-6.
        window: int (optional)
            Window around the jump, where the voltage is evaluated and classified as 'superconducting branch'. Default
            is 5 that considers two values below and 2 values above the jump.

        Returns
        -------
        I_cs: numpy.array
            Critical current values.
        I_rs: numpy.array (optional)
            Retrapping current values.
        properties: dict
            Properties of all found peaks (not only I_c and I_r, but also further jumps), such as corresponding currents,
            voltages, differential resistances, indices as well as returns of the used peak finding algorithm.
        """

        def get_fwhm(y, peak, mask):
            """ Get bounds of full width at half maximum """
            left, right = peak[1]['left_bases'][mask].item(), peak[1]['right_bases'][mask].item()
            y = y[slice(left, right)] - np.nanmean(y)  # use left and right bases as bounds and subtract floor (mean(y))
            hm = np.max(y) / 2  # half maximum
            max_idx = np.argmax(y)  # index of maximum
            d = np.sign(hm - y[:-1]) - np.sign(hm - y[1:])
            left_idx = np.argwhere(d[:max_idx] > 0)[-1]
            right_idx = max_idx + np.argwhere(d[max_idx:] < 0)[0] + 1
            return np.concatenate((left_idx, right_idx)) + left

        Y_name = [key for key, val in sys._getframe().f_back.f_locals.items() if np.array_equal(val, Y)][0]
        if self.V_offset is None and self.sweeptype in [0, 1]:  # voltage offset to identify superconducting branch
            self.get_offset(x=V, y=I)
        if self.sweeptype == 0:  # halfswing
            V_c, I_c, peaks_c = np.copy(V), np.copy(I), np.copy(peaks)
            V_r, I_r, peaks_r = np.copy(V), np.copy(I), np.copy(peaks)
        elif self.sweeptype == 1:  # 4quadrants
            V_c, I_c, peaks_c = np.copy(V[::2]), np.copy(I[::2]), np.copy(peaks[::2])
            V_r, I_r, peaks_r = np.copy(V[1::2]), np.copy(I[1::2]), np.copy(peaks[1::2])
        else:  # custom sweep type
            def f(ind1D, prop1D, I1D, V1D, Y1D):
                try:
                    return {**prop1D,
                            **{k: v for k, v in zip(('I', 'V', Y_name, 'index'),
                                                    (I1D[ind1D], V1D[ind1D], Y1D[ind1D], ind1D))}}
                except IndexError:  # if no peak found return np.nan
                    return {k: v for k, v in zip(('I', 'V', Y_name, 'index'),
                                                 (np.array([np.nan]), np.array([np.nan]), np.array([np.nan]),
                                                  np.array([np.nan])))}

            if len(V.shape) - 1 == 1:
                return np.array(list(map(lambda ind1D, prop1D, I1D, V1D, Y1D:
                                         f(ind1D, prop1D, I1D, V1D, Y1D),
                                         *list(zip(*peaks)), I, V, Y)))
            elif len(V.shape) - 1 == 2:
                return np.array(list(map(lambda peaks2D, I2D, V2D, Y2D:
                                         list(map(lambda ind1D, prop1D, I1D, V1D, Y1D:
                                                  f(ind1D, prop1D, I1D, V1D, Y1D),
                                                  *list(zip(*peaks2D)), I2D, V2D, Y2D)),
                                         peaks, I, V, Y)))
            elif len(V.shape) - 1 == 3:
                return np.array(list(map(lambda peaks3D, I3D, V3D, Y3D:
                                         list(map(lambda peaks2D, I2D, V2D, Y2D:
                                                  list(map(lambda ind1D, prop1D, I1D, V1D, Y1D:
                                                           f(ind1D, prop1D, I1D, V1D, Y1D),
                                                           *list(zip(*peaks2D)), I2D, V2D, Y2D)),
                                                  peaks3D, I3D, V3D, Y3D)),
                                         peaks, I, V, Y)))
            else:
                raise NotImplementedError('No algorithm implemented for custom sweep type and scan_dim > 3.')
        if len(V.shape) - 1 == 1:
            ''' critical current '''
            masks_c = list(map(lambda V_c1D, peak1D:
                               np.logical_and(np.logical_and((-tol_offset + self.V_offset) <= np.mean(
                                   V_c1D[peak1D[0] - np.tile([np.arange(int(window) // 2) + 1], [len(peak1D[0]), 1]).T],
                                   axis=0),
                                                             np.mean(V_c1D[peak1D[0] - np.tile(
                                                                 [np.arange(int(window) // 2) + 1],
                                                                 [len(peak1D[0]), 1]).T], axis=0) <= (
                                                                     +tol_offset + self.V_offset)),
                                              np.logical_not(np.logical_and((-tol_offset + self.V_offset) <= np.mean(
                                                  V_c1D[peak1D[0] + np.tile([np.arange(int(window) // 2) + 1],
                                                                            [len(peak1D[0]), 1]).T], axis=0),
                                                                            np.mean(V_c1D[peak1D[0] + np.tile(
                                                                                [np.arange(int(window) // 2) + 1],
                                                                                [len(peak1D[0]), 1]).T], axis=0) <= (
                                                                                    +tol_offset + self.V_offset)))),
                               V_c, peaks_c))
            if not np.all(np.any(masks_c, axis=1)):
                for i in np.argwhere(~np.any(masks_c, axis=1)):
                    masks_c[i.item()][1] = True
            # I_cs = np.array(list(map(lambda I_c1D, peak1D, masks_c1D:
            #                         I_c1D[peak1D[0][masks_c1D][0]],
            #                         I_c, peaks_c, masks_c)),
            #                dtype=float)
            I_cs = np.array(list(map(lambda I_c1D, Y1D, peak1D, masks_c1D:
                                     ufloat(nominal_value=I_c1D[peak1D[0][masks_c1D][0]],
                                            std_dev=np.abs(
                                                np.diff(I_c1D[get_fwhm(Y1D, peak1D, masks_c1D)])).item() / 2),
                                     # half width half maximum
                                     I_c, Y, peaks_c, masks_c)),
                            dtype=object)
            ''' retrapping current '''
            masks_r = list(map(lambda V_r1D, peak1D:
                               np.logical_and(np.logical_not(np.logical_and((-tol_offset + self.V_offset) <= np.mean(
                                   V_r1D[peak1D[0] - np.tile([np.arange(int(window) // 2) + 1], [len(peak1D[0]), 1]).T],
                                   axis=0),
                                                                            np.mean(V_r1D[peak1D[0] - np.tile(
                                                                                [np.arange(int(window) // 2) + 1],
                                                                                [len(peak1D[0]), 1]).T], axis=0) <= (
                                                                                    +tol_offset + self.V_offset))),
                                              np.logical_and((-tol_offset + self.V_offset) <= np.mean(V_r1D[peak1D[
                                                                                                                0] + np.tile(
                                                  [np.arange(int(window) // 2) + 1], [len(peak1D[0]), 1]).T], axis=0),
                                                             np.mean(V_r1D[peak1D[0] + np.tile(
                                                                 [np.arange(int(window) // 2) + 1],
                                                                 [len(peak1D[0]), 1]).T], axis=0) <= (
                                                                     +tol_offset + self.V_offset))),
                               V_r, peaks_r))
            if not np.all(np.any(masks_r, axis=1)):
                for i in np.argwhere(~np.any(masks_r, axis=1)):
                    masks_r[i.item()][0] = True
            # I_rs = np.array(list(map(lambda I_r1D, peak1D, masks_r1D:
            #                         I_r1D[peak1D[0][masks_r1D][0]+1],
            #                         I_r, peaks_r, masks_r)),
            #                dtype=float)
            I_rs = np.array(list(map(lambda I_r1D, Y1D, peak1D, masks_r1D:
                                     ufloat(nominal_value=I_r1D[peak1D[0][masks_r1D][0]],
                                            std_dev=np.abs(
                                                np.diff(I_r1D[get_fwhm(Y1D, peak1D, masks_r1D)])).item() / 2),
                                     # half width half maximum
                                     I_r, Y, peaks_r, masks_r)),
                            dtype=object)
            ''' properties '''
            properties = np.array(list(map(lambda ind1D, prop1D, I1D, V1D, Y1D:
                                           {**prop1D,
                                            **{k: v for k, v in zip(('I', 'V', Y_name, 'index'),
                                                                    (I1D[ind1D], V1D[ind1D], Y1D[ind1D], ind1D))}},
                                           *list(zip(*peaks)), I, V, Y)))
            ''' gap voltage '''
            if Vg:
                slcs_g = list(map(lambda j, m:
                                  slice(-peaks[int(not j)][0][m[0]].item() - 1, peaks[j][0][m[1]].item()),
                                  *zip(*enumerate(zip(*(masks_c, masks_r))))))
                V_gs = unp.uarray(nominal_values=[np.nanmean(V[j][slc]) for j, slc in enumerate(slcs_g)],
                                  std_devs=[np.nansem(V[j][slc]) for j, slc in enumerate(slcs_g)])
        elif len(V.shape) - 1 == 2:
            ''' critical current '''
            masks_c = map(lambda V_c2D, peaks_c2D:
                          map(lambda V_c1D, peak1D: np.logical_and(np.logical_and(
                              (-tol_offset + self.V_offset) <= np.mean(
                                  V_c1D[peak1D[0] - np.tile([np.arange(int(window) // 2) + 1], [len(peak1D[0]), 1]).T],
                                  axis=0),
                              np.mean(
                                  V_c1D[peak1D[0] - np.tile([np.arange(int(window) // 2) + 1], [len(peak1D[0]), 1]).T],
                                  axis=0) <= (+tol_offset + self.V_offset)),
                              np.logical_not(np.logical_and(
                                  (-tol_offset + self.V_offset) <= np.mean(V_c1D[
                                                                               peak1D[
                                                                                   0] + np.tile(
                                                                                   [
                                                                                       np.arange(
                                                                                           int(
                                                                                               window) // 2) + 1],
                                                                                   [
                                                                                       len(
                                                                                           peak1D[
                                                                                               0]),
                                                                                       1]).T],
                                                                           axis=0),
                                  np.mean(V_c1D[peak1D[0] + np.tile(
                                      [np.arange(int(window) // 2) + 1],
                                      [len(peak1D[0]), 1]).T], axis=0) <= (
                                          +tol_offset + self.V_offset)))),
                              V_c2D, peaks_c2D),
                          V_c, peaks_c)
            # I_cs = np.array(list(map(lambda I_c2D, peaks_c2D, masks_c2D:
            #                         list(map(lambda I_c1D, peak1D, masks_c1D:
            #                                  I_c1D[peak1D[0][masks_c1D][0]],
            #                                  I_c2D, peaks_c2D, masks_c2D)),
            #                         I_c, peaks_c, masks_c)),
            #                dtype=float)
            I_cs = np.array(list(map(lambda I_c2D, Y2D, peaks_c2D, masks_c2D:
                                     list(map(lambda I_c1D, Y1D, peak1D, masks_c1D:
                                              ufloat(nominal_value=I_c1D[peak1D[0][masks_c1D][0]],
                                                     std_dev=np.abs(
                                                         np.diff(I_c1D[get_fwhm(Y1D, peak1D, masks_c1D)])).item() / 2),
                                              # half width half maximum
                                              I_c2D, Y2D, peaks_c2D, masks_c2D)),
                                     I_c, Y, peaks_c, masks_c)),
                            dtype=object)
            ''' retrapping current '''
            masks_r = map(lambda V_r2D, peaks_r2D:
                          map(lambda V_r1D, peak1D: np.logical_and(np.logical_not(np.logical_and(
                              (-tol_offset + self.V_offset) <= np.mean(
                                  V_r1D[peak1D[0] - np.tile([np.arange(int(window) // 2) + 1], [len(peak1D[0]), 1]).T],
                                  axis=0),
                              np.mean(
                                  V_r1D[peak1D[0] - np.tile([np.arange(int(window) // 2) + 1], [len(peak1D[0]), 1]).T],
                                  axis=0) <= (+tol_offset + self.V_offset))),
                              np.logical_and(
                                  (-tol_offset + self.V_offset) <= np.mean(V_r1D[
                                                                               peak1D[
                                                                                   0] + np.tile(
                                                                                   [
                                                                                       np.arange(
                                                                                           int(
                                                                                               window) // 2) + 1],
                                                                                   [
                                                                                       len(
                                                                                           peak1D[
                                                                                               0]),
                                                                                       1]).T],
                                                                           axis=0),
                                  np.mean(V_r1D[peak1D[0] + np.tile(
                                      [np.arange(int(window) // 2) + 1],
                                      [len(peak1D[0]), 1]).T], axis=0) <= (
                                          +tol_offset + self.V_offset))),
                              V_r2D, peaks_r2D),
                          V_r, peaks_r)
            # I_rs = np.array(list(map(lambda I_r2D, peaks_r2D, masks_r2D:
            #                         list(map(lambda I_r1D, peaks_r1D, masks_r1D:
            #                                  I_r1D[peaks_r1D[0][masks_r1D][0]+1],
            #                                  I_r2D, peaks_r2D, masks_r2D)),
            #                         I_r, peaks_r, masks_r)),
            #                dtype=float)
            I_rs = np.array(list(map(lambda I_r2D, Y2D, peaks_r2D, masks_r2D:
                                     list(map(lambda I_r1D, Y1D, peak1D, masks_r1D:
                                              ufloat(nominal_value=I_r1D[peak1D[0][masks_r1D][0]],
                                                     std_dev=np.abs(
                                                         np.diff(I_r1D[get_fwhm(Y1D, peak1D, masks_r1D)])).item() / 2),
                                              # half width half maximum
                                              I_r2D, Y2D, peaks_r2D, masks_r2D)),
                                     I_r, Y, peaks_r, masks_r)),
                            dtype=object)
            ''' properties '''
            properties = np.array(list(map(lambda peaks2D, I2D, V2D, Y2D:
                                           list(map(lambda ind1D, prop1D, I1D, V1D, Y1D:
                                                    {**prop1D,
                                                     **{k: v for k, v in zip(('I', 'V', Y_name, 'index'),
                                                                             (I1D[ind1D], V1D[ind1D], Y1D[ind1D],
                                                                              ind1D))}},
                                                    *list(zip(*peaks2D)), I2D, V2D, Y2D)),
                                           peaks, I, V, Y)))
        elif len(V.shape) - 1 == 3:
            ''' critical current '''
            masks_c = map(lambda V_c3D, peaks_c3D:
                          map(lambda V_c2D, peaks_c2D:
                              map(lambda V_c1D, peak1D: np.logical_and(np.logical_and(
                                  (-tol_offset + self.V_offset) <= np.mean(V_c1D[peak1D[0] - np.tile(
                                      [np.arange(int(window) // 2) + 1], [len(peak1D[0]), 1]).T], axis=0),
                                  np.mean(V_c1D[peak1D[0] - np.tile([np.arange(int(window) // 2) + 1],
                                                                    [len(peak1D[0]), 1]).T], axis=0) <= (
                                          +tol_offset + self.V_offset)),
                                  np.logical_not(np.logical_and(
                                      (-tol_offset + self.V_offset) <= np.mean(
                                          V_c1D[peak1D[0] + np.tile(
                                              [np.arange(int(window) // 2) + 1],
                                              [len(peak1D[0]), 1]).T], axis=0),
                                      np.mean(V_c1D[peak1D[0] + np.tile(
                                          [np.arange(int(window) // 2) + 1],
                                          [len(peak1D[0]), 1]).T], axis=0) <= (
                                              +tol_offset + self.V_offset)))),
                                  V_c2D, peaks_c2D),
                              V_c3D, peaks_c3D),
                          V_c, peaks_c)
            # I_cs = np.array(list(map(lambda I_c3D, peaks_c3D, masks_c3D:
            #                         list(map(lambda I_c2D, peaks_c2D, masks_c2D:
            #                                  list(map(lambda I_c1D, peak1D, masks_c1D:
            #                                           I_c1D[peak1D[0][masks_c1D][0]],
            #                                           I_c2D, peaks_c2D, masks_c2D)),
            #                                  I_c3D, peaks_c3D, masks_c3D)),
            #                         I_c, peaks_c, masks_c)),
            #                dtype=float)
            I_cs = np.array(list(map(lambda I_c3D, Y3D, peaks_c3D, masks_c3D:
                                     list(map(lambda I_c2D, Y2D, peaks_c2D, masks_c2D:
                                              list(map(lambda I_c1D, Y1D, peak1D, masks_c1D:
                                                       ufloat(nominal_value=I_c1D[peak1D[0][masks_c1D][0]],
                                                              std_dev=np.abs(np.diff(
                                                                  I_c1D[get_fwhm(Y1D, peak1D, masks_c1D)])).item() / 2),
                                                       # half width half maximum
                                                       I_c2D, Y2D, peaks_c2D, masks_c2D)),
                                              I_c3D, Y3D, peaks_c3D, masks_c3D)),
                                     I_c, Y, peaks_c, masks_c)),
                            dtype=object)
            ''' retrapping current '''
            masks_r = map(lambda V_r3D, peaks_r3D:
                          map(lambda V_r2D, peaks_r2D:
                              map(lambda V_r1D, peak1D: np.logical_and(np.logical_not(np.logical_and(
                                  (-tol_offset + self.V_offset) <= np.mean(V_r1D[peak1D[0] - np.tile(
                                      [np.arange(int(window) // 2) + 1], [len(peak1D[0]), 1]).T], axis=0),
                                  np.mean(V_r1D[peak1D[0] - np.tile([np.arange(int(window) // 2) + 1],
                                                                    [len(peak1D[0]), 1]).T], axis=0) <= (
                                          +tol_offset + self.V_offset))),
                                  np.logical_and(
                                      (-tol_offset + self.V_offset) <= np.mean(
                                          V_r1D[peak1D[0] + np.tile(
                                              [np.arange(int(window) // 2) + 1],
                                              [len(peak1D[0]), 1]).T], axis=0),
                                      np.mean(V_r1D[peak1D[0] + np.tile(
                                          [np.arange(int(window) // 2) + 1],
                                          [len(peak1D[0]), 1]).T], axis=0) <= (
                                              +tol_offset + self.V_offset))),
                                  V_r2D, peaks_r2D),
                              V_r3D, peaks_r3D),
                          V_r, peaks_r)
            # I_rs = np.array(list(map(lambda I_r3D, peaks_r3D, masks_r3D:
            #                         list(map(lambda I_r2D, peaks_r2D, masks_r2D:
            #                                  list(map(lambda I_r1D, peak1D, masks_r1D:
            #                                           I_r1D[peak1D[0][masks_r1D][0]],
            #                                           I_r2D, peaks_r2D, masks_r2D)),
            #                                  I_r3D, peaks_r3D, masks_r3D)),
            #                         I_r, peaks_r, masks_r)),
            #                dtype=float)
            I_rs = np.array(list(map(lambda I_r3D, Y3D, peaks_r3D, masks_r3D:
                                     list(map(lambda I_r2D, Y2D, peaks_r2D, masks_r2D:
                                              list(map(lambda I_r1D, Y1D, peak1D, masks_r1D:
                                                       ufloat(nominal_value=I_r1D[peak1D[0][masks_r1D][0]],
                                                              std_dev=np.abs(np.diff(
                                                                  I_r1D[get_fwhm(Y1D, peak1D, masks_r1D)])).item() / 2),
                                                       # half width half maximum
                                                       I_r2D, Y2D, peaks_r2D, masks_r2D)),
                                              I_r3D, Y3D, peaks_r3D, masks_r3D)),
                                     I_r, Y, peaks_r, masks_r)),
                            dtype=object)
            properties = np.array(list(map(lambda peaks3D, I3D, V3D, Y3D:
                                           list(map(lambda peaks2D, I2D, V2D, Y2D:
                                                    list(map(lambda ind1D, prop1D, I1D, V1D, Y1D:
                                                             {**prop1D,
                                                              **{k: v for k, v in zip(('I', 'V', Y_name, 'index'),
                                                                                      (I1D[ind1D], V1D[ind1D],
                                                                                       Y1D[ind1D], ind1D))}},
                                                             *list(zip(*peaks2D)), I2D, V2D, Y2D)),
                                                    peaks3D, I3D, V3D, Y3D)),
                                           peaks, I, V, Y)))
        else:
            raise ValueError('Scan dimension must be in {1, 2, 3}')
        """
        ans = {'I_c': I_cs, 'prop': properties}
        if Ir:
            ans['I_r'] = I_rs
        if Vg:
            ans['V_g': V_gs]
        return dict2obj(ans)
        """
        if Ir:
            if Vg:
                return I_cs, I_rs, V_gs, properties
            else:
                return I_cs, I_rs, properties
        else:
            if Vg:
                return I_cs, V_gs, properties
            else:
                return I_cs, properties

    def get_Vg(self, V=None, binwidth=2e-6, subgap=False):
        """
        Gets gap voltages. Therefore, all measured voltages are collected in bins with bin width <binwidth> and the 3
        most prominent peaks are interpreted as superconducting branch (2nd) and gap (1st and 3rd). Each peak is fitted
        to a Gaussian distribution using a fine histogram ranging from `left_bases` to `right_bases` obtained from the
        peak finder `scipy.signal.find_peaks`. The returned gap voltage equals the mean ± std of the differences of the
        3 peaks.
        If wanted, the sub-gap voltage is calculated as maximal absolute voltage between two neighboring peaks, where
        outliers are removed by considering only the median ± median absolute deviation. The returned sub-gap voltage
        equals the maximal absolute voltage and the error is estimated as difference of the two maximal absolute voltages.

        Parameters
        ----------
        V: numpy.array (optional)
            An N-dimensional array containing voltage values. Default is None that means <self.V>.
        binwidth: float (optional)
            Histograms bin width in Volts. Default is 2e-6.
        subgap: bool (optional)
            Condition, if sub-gap voltages are returned, too. Default is False.

        Returns
        -------
        V_g: numpy.array
            Gap voltage.
        V_sg: numpy.array (optional)
            Sub-gap voltage.
        """

        def gaussian(x, A, mu, sigma, k):
            return A * np.exp(-0.5 * ((x - mu) / sigma) ** 2) + k

        if V is None:
            V = self.V
        ''' prepare numpy array of voltage values to handle different scan-dimensions '''
        if len(V.shape) - 1 == 1:
            _V = V[:, np.newaxis, np.newaxis, :]  # insert dummy x- and y-axis
        elif len(V.shape) - 1 == 2:
            _V = V[:, :, np.newaxis, :]  # insert dummy y-axis
        elif len(V.shape) - 1 == 3:
            _V = V
        else:
            raise ValueError('Scan dimension must be in {1, 2, 3}')
        ''' find gap (and sub-gap) voltages '''
        nop = 3  # number of peaks to find
        V_g = np.ones(shape=(*_V.shape[1:3], nop), dtype=object) * np.nan
        V_sg = np.ones(shape=(*_V.shape[1:3], nop - 1), dtype=object) * np.nan
        for x in range(_V.shape[1]):  # iterate x-values
            for y in range(_V.shape[2]):  # iterate y-values
                ''' raw histogram '''
                p, edges = np.histogram(np.concatenate(_V[:, x, y]),
                                        bins=int((np.nanmax(_V[:, x, y]) - np.nanmin(_V[:, x, y])) / binwidth))
                # find <nop> peaks
                prominence = np.max(p)
                while True:
                    peaks = sig.find_peaks(p, prominence=prominence)
                    if peaks[0].size < nop:
                        prominence /= 2
                    elif peaks[0].size == nop:
                        break
                    else:
                        raise ValueError('V_g: Too many peaks found. Probably adjust bin width!')
                ''' fine histogram '''
                for i in range(nop):
                    p1, edges1 = np.histogram(np.concatenate(_V[:, x, y]),
                                              bins=100,
                                              range=(edges[peaks[1]['left_bases'][i]],
                                                     edges[peaks[1]['right_bases'][i] + 1]))
                    bins1 = np.convolve(edges1, np.ones((2,)) / 2, mode='valid')
                    p0 = [np.max(p1),
                          np.median(list(filter(lambda val: edges[peaks[1]['left_bases'][i]] <= val <= edges[
                              peaks[1]['right_bases'][i] + 1],
                                                np.concatenate(_V[:, x, y])))),
                          np.std(list(filter(lambda val: edges[peaks[1]['left_bases'][i]] <= val <= edges[
                              peaks[1]['right_bases'][i] + 1],
                                             np.concatenate(_V[:, x, y])))),
                          0.]
                    popt, pcov = opt.curve_fit(gaussian, bins1, p1, p0=p0, maxfev=1000)
                    V_g[x, y, i] = ufloat(nominal_value=popt[1],
                                          std_dev=np.sqrt(pcov[1, 1]) + np.abs(popt[2]))
                if subgap:
                    for i in range(nop - 1):
                        sg = np.array(list(filter(lambda val: edges[peaks[1]['right_bases'][i]] <= val <= edges[
                            peaks[1]['left_bases'][i + 1] + 1],
                                                  np.concatenate(_V[:, x, y]))))
                        # find outliers
                        sg_median = ufloat(nominal_value=np.median(sg),
                                           std_dev=stats.median_abs_deviation(sg))
                        V_sg[x, y, i] = ufloat(
                            nominal_value=[np.nanmin, np.nanmax][i](sg[np.logical_and(sg_median.n - sg_median.s < sg,
                                                                                      sg < sg_median.n + sg_median.s)]),
                            std_dev=np.abs(np.diff(sorted(sg[np.logical_and(sg_median.n - sg_median.s < sg,
                                                                            sg < sg_median.n + sg_median.s)],
                                                          reverse=~bool(i))[-2:]).item()))
        if subgap:
            return np.squeeze(V_g), np.squeeze(V_sg)
        else:
            return np.squeeze(V_g)

    class switching_current(object):
        """
        This is an analysis class for switching current measurements.
        """
        def __init__(self, sweeps, settings):
            self.sweeps = sweeps
            self.settings = settings

            self.P = self.P_fit = self.edges = self.edges_fit = self.bins = self.bins_fit = None
            self.Delta_I = self.Delta_I_fit = self.dIdt = None
            self.Gamma = self.Gamma_fit = self.x = self.x_fit = self.y = self.y_fit = None
            self.popt = self.pcov = self.fit_res = self.I_c = self.omega_0 = self.T_esc = None
            self.fig = self.ax1 = self.ax2 = self.ax3 = None

        def fit(self, I_0, omega_0, dIdt=None, **kwargs):
            """
            Creates switching current histogram, calculates and escape rate and recalculates the fit to the switching
            current distribution.

            Parameters
            ----------
            I_0: array-like
                Switching currents.
            omega_0: float
                Plasma frequency used for fit.
            dIdt: float (optional)
                Sweep rate (in A/s). Default is sweeps step width * nplc / plc.
            kwargs:
                Keyword arguments forwarded to numpy.histogram. Defaults are
                * bins=10,
                * range=(min(<I_0>), max(<I_0>)),
                * normed=None,
                * weights=None,
                * density=None.

            Returns
            -------
            P: np.array
                Probability distribution
            Gamma: np.array
                Escape rate
            popt: list
                Optimal fit parameters
            pcov: list
                Covariance matrix
            I_c: float
                Fitted critical current

            Examples
            --------
            >>> ivc.scm.fit(I_0=I_0*1e6, omega_0=14e9, bins=50);
            """
            self.omega_0 = omega_0
            ''' histogram '''
            if 'range' not in kwargs:
                kwargs['range'] = (np.nanmin(I_0), np.nanmax(I_0))
            if 'weights' not in kwargs:
                kwargs['weights'] = np.ones_like(I_0) / I_0.size  # normed to 1
            with np.errstate(invalid='ignore'):
                self.P, self.edges = np.histogram(a=I_0, **kwargs)
            self.bins = np.convolve(self.edges, np.ones((2,)) / 2,
                                    mode='valid')  # center of bins by moving average with window length 2
            ''' escape rate '''
            self.Delta_I = np.nanmean(np.diff(self.edges))
            if dIdt is None:
                self.dIdt = self.sweeps[0][2] * np.squeeze(np.unique(self.settings.IVD.sense_nplc)) / self.settings.IVD.plc
            else:
                self.dIdt = dIdt
            with np.errstate(divide='ignore'):  # raises warning due to divide by zero
                self.Gamma = self.dIdt / self.Delta_I * np.array([np.log(np.sum(self.P[k:]) / np.sum(self.P[k + 1:])) for k, _ in enumerate(self.P)])
            ''' fit norm. escape rate '''
            self.x = self.bins
            with np.errstate(divide='ignore'):  # raises warning due divide by zero
                self.y = np.log(unp.nominal_values(self.omega_0) / (2 * np.pi * self.Gamma)) ** (2 / 3)
            self.popt, self.pcov = np.polyfit(x=self.x[np.isfinite(self.y)],
                                              y=self.y[np.isfinite(self.y)],
                                              cov=True,
                                              deg=1)
            self.fit_res = unp.uarray(nominal_values=self.popt,
                                      std_devs=np.sqrt(np.diagonal(self.pcov)))
            self.I_c = -self.fit_res[1] / self.fit_res[0]
            self.T_esc = -4 * np.sqrt(2) * const.Phi_0 / (3 * 2 * np.pi * const.k) / (self.fit_res[0] * unp.sqrt(self.fit_res[1]))
            return dict2obj({'P': self.P,
                             'Gamma': self.Gamma,
                             'popt': self.popt,
                             'pcov': self.pcov,
                             'I_c': self.I_c,
                             'T_esc': self.T_esc,
                             })

        def plot(self,
                 xlabel='escape current $ I_{esc}~(A) $',
                 y1label='escape probability $ P $',
                 y2label='escape rate $ \Gamma_{esc}~(s^{-1}) $',
                 y3label='norm. escape rate $ \ln(\omega_0/2\pi\Gamma) $',
                 errors=False,
                 alpha=1,
                 figsize=(6, 6)):
            """
            Plots switching current histogram, escape rate and normalized escape rate as well as their fits.

            Parameters
            ----------
            xlabel: str
                X-axis label. Default is 'escape current $ I_{esc}~(A) $'.
            y1label: str
                Y-axis label of histogram. Default is 'escape probability $ P $'.
            y2label: str
                Y-axis label of escape rate. Default is 'escape rate $ \Gamma_{esc}~(s^{-1}) $'.
            y3label: str
                Y-axis label of normalized escape rate. Default is 'norm. escape rate $ \ln(\omega_0/2\pi\Gamma) $'.
            errors: bool
                If this is set to True, error bands are plotted.
            alpha: int
                Multiplier for nop of fit. Default is 1, which uses histogram bins for the fit.
                Note that alpha > 1 shifts the fit in x-direction for unknown reasons, which is corrected in the plot
            figsize : (float, float)
                Width, height in inches. Default is :rc:`figure.figsize`


            Returns
            -------
            fig: matplotlib.figure.Figure
                Plot figure, which shows the data and the fit of the  switching current distribution.
            ax: matplotlib.axes._subplots.AxesSubplot array of Axes objects
                ax[0], ax[1] and ax[2] contains switching current distribution, escape rate and normalized escape rate,
                respectively.

            Examples
            --------
            >>> ivc.scm.plot()
            """
            ''' calculate fitted escape rate and fitted switching current distribution '''
            self.edges_fit, self.Delta_I_fit = np.linspace(np.min(self.edges), np.max(self.edges), self.bins.size * alpha + 1, retstep=True)
            self.bins_fit = np.convolve(self.edges_fit, np.ones((2,)) / 2, mode='valid')
            self.x_fit = self.bins_fit
            if errors:  # this works also for errors=True, but is much slower
                self.y_fit = self.fit_res[0] * self.x_fit + self.fit_res[1]
                self.Gamma_fit = self.omega_0 / (2 * np.pi) * unp.exp(-self.y_fit ** (3 / 2))
                self.P_fit = np.array([self.Gamma_fit[k] / self.dIdt * unp.exp(
                    -np.sum(self.Gamma_fit[:k + 1]) * self.Delta_I_fit / self.dIdt) for k, _ in enumerate(self.x_fit)])
            else:
                self.y_fit = self.popt[0] * self.x_fit + self.popt[1]
                self.Gamma_fit = unp.nominal_values(self.omega_0) / (2 * np.pi) * np.exp(-self.y_fit ** (3 / 2))
                self.P_fit = np.array([self.Gamma_fit[k] / self.dIdt * np.exp(
                    -np.sum(self.Gamma_fit[:k + 1]) * self.Delta_I_fit / self.dIdt) for k, _ in enumerate(self.x_fit)])
            self.P_fit /= np.sum(self.P_fit) / alpha
            ''' plot '''
            self.fig, (self.ax1, self.ax2) = plt.subplots(nrows=2,
                                                          ncols=1,
                                                          sharex=True,
                                                          figsize=figsize)
            self.ax3 = plt.twinx(self.ax2)
            self.ax2.set_xlabel(xlabel)
            self.ax1.set_ylabel(y1label)
            self.ax2.set_ylabel(y2label)
            self.ax3.set_ylabel(y3label)
            ''' histogram '''
            lgd11 = self.ax1.bar(x=self.bins,
                                 height=self.P,
                                 width=np.mean(np.gradient(self.edges)))
            # self.ax1.plot(self.x_fit,
            #              self.P_fit,
            #              label='fit',
            #              ls='-',
            #              color='black')
            lgd12, = self.ax1.plot(self.x_fit - self.Delta_I / 2,
                                   unp.nominal_values(self.P_fit),
                                   ls='-',
                                   color='black')
            if errors:
                self.ax1.fill_between(x=self.x_fit,
                                      y1=(unp.nominal_values(self.P_fit) - unp.std_devs(self.P_fit)).clip(0),
                                      # replace negative values by 0
                                      y2=unp.nominal_values(self.P_fit) + unp.std_devs(self.P_fit),
                                      color='black', alpha=0.2)
                lgd13, = self.ax1.fill(np.NaN, np.NaN, 'black', alpha=0.2)  # dummy for legend
                self.ax1.legend(((lgd11,), (lgd12, lgd13)), ('data', 'fit'), loc='upper left')
            else:
                self.ax1.legend(((lgd11,), (lgd12,)), ('data', 'fit'), loc='upper left')
            # self.ax1.legend(loc='upper left')
            ''' escape rate '''
            lgd21, = self.ax2.semilogy(self.x,
                                       self.Gamma,
                                       'b.',
                                       label='$ \Gamma $')
            # lgd22, = self.ax2.semilogy(self.x_fit,
            #                           self.Gamma_fit,
            #                           'b-',
            #                           label='fit')
            lgd22, = self.ax2.semilogy(self.x_fit,
                                       unp.nominal_values(self.Gamma_fit),
                                       'b-',
                                       label='fit')
            if errors:
                self.ax2.fill_between(x=self.x_fit,
                                      y1=unp.nominal_values(self.Gamma_fit) - unp.std_devs(self.Gamma_fit),
                                      y2=unp.nominal_values(self.Gamma_fit) + unp.std_devs(self.Gamma_fit),
                                      color='b', alpha=0.2)
                # lgd23, = self.ax2.fill(np.NaN, np.NaN, 'b', alpha=0.2)  # dummy for legend
            ''' normalized escape rate '''
            lgd31, = self.ax3.plot(self.x,
                                   self.y,
                                   'r.',
                                   label='$ \ln(\omega_0/2\pi\Gamma) $')
            # lgd32, = self.ax3.plot(self.x_fit,
            #                       self.y_fit,
            #                       'r-',
            #                       label='fit')
            lgd32, = self.ax3.plot(self.x_fit,
                                   unp.nominal_values(self.y_fit),
                                   'r-',
                                   label='fit')
            if errors:
                pass
                self.ax3.fill_between(x=self.x_fit,
                                      y1=unp.nominal_values(self.y_fit) - unp.std_devs(self.y_fit),
                                      y2=unp.nominal_values(self.y_fit) + unp.std_devs(self.y_fit),
                                      color='r', alpha=0.2)
                # lgd33, = self.ax3.fill(np.NaN, np.NaN, 'r', alpha=0.2)  # dummy for legend
                # self.ax2.legend(((lgd21, lgd22, lgd23), (lgd31, lgd32, lgd33)), ('$ \Gamma $', '$ \ln(\omega_0/2\pi\Gamma) $'), loc='upper center')
            else:
                self.ax2.legend(((lgd21, lgd22), (lgd31, lgd32)), ('$ \Gamma $', '$ \ln(\omega_0/2\pi\Gamma) $'),
                                loc='upper center')
            plt.subplots_adjust(hspace=0)
            plt.show()

            return self.fig, (self.ax1, self.ax2, self.ax3)



# -*- coding: utf-8 -*-
"""
Start remote interface service (RIS)
@author: HR@KIT/2018
"""
import qkit
import logging


def _load_ri_service():
    logging.info(__file__+": loading remote interface service")
    from qkit.core.lib.com.ri_service import RISThread
    qkit.ris = RISThread()

if qkit.cfg.get('load_ri_service',False):
    qkit.cfg['load_ri_service']=True
    _load_ri_service()
else:
    qkit.cfg['load_ri_service']=False


# OxfordInstruments_Kelvinox_IGH.py class, to perform the communication between the Wrapper and the device
# Guenevere Prawiroatmodjo <guen@vvtp.tudelft.nl>, 2009
# Pieter de Groot <pieterdegroot@gmail.com>, 2009
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
from time import time, sleep
from qkit import visa
import types
import logging

class OxfordInstruments_Kelvinox_IGH(Instrument):
    '''
    This is the python driver for the Oxford Instruments Kelvinox IGH Dilution Refrigerator and
    Intelligent Dilution Refrigerator Power Supply (IDR PS).

    Usage:
    Initialize with
    <name> = instruments.create('name', 'OxfordInstruments_Kelvinox_IGH', address='<Instrument address>')
    <Instrument address> = ASRL1::INSTR

    Note: Since the ISOBUS allows for several instruments to be managed in parallel, the command
    which is sent to the device starts with '@n', where n is the ISOBUS instrument number.

    '''
#TODO: auto update script
#TODO: get doesn't always update the wrapper! (e.g. when input is an int and output is a string)

    def __init__(self, name, address, number=5):
        '''
        Initializes the Oxford Instruments Kelvinox IGH Dilution Refrigerator.

        Input:
            name (string)    : name of the instrument
            address (string) : instrument address
            number (int)     : ISOBUS instrument number

        Output:
            None
        '''
        logging.debug(__name__ + ' : Initializing instrument')
        Instrument.__init__(self, name, tags=['physical'])


        self._address = address
        self._number = number
        self._visainstrument = visa.SerialInstrument(self._address)
        self._values = {}
        self._visainstrument.stop_bits = 2

        #Add parameters
        self.add_parameter('1K_pot_temp', type=float,
            flags=Instrument.FLAG_GET)
        self.add_parameter('mix_chamber_temp', type=float,
            flags=Instrument.FLAG_GET)

        # Add functions
        self.add_function('get_all')
        self.get_all()

    def get_all(self):
        '''
        Reads all implemented parameters from the instrument,
        and updates the wrapper.

        Input:
            None

        Output:
            None
        '''
        logging.info(__name__ + ' : reading all settings from instrument')
        self.get_1K_pot_temp()
        self.get_mix_chamber_temp()

    # Functions
    def _execute(self, message):
        '''
        Write a command to the device

        Input:
            message (str) : write command for the device

        Output:
            None
        '''
        logging.info(__name__ + ' : Send the following command to the device: %s' % message)
        self._visainstrument.write('@%s%s' % (self._number, message))
        sleep(20e-3) # wait for the device to be able to respond
        result = self._visainstrument.read()
        if result.find('?') >= 0:
            print "Error: Command %s not recognized" % message
        else:
            return result

    def identify(self):
        '''
        Identify the device

        Input:
            None

        Output:
            None
        '''
        logging.info(__name__ + ' : Identify the device')
        return self._execute('V')

    def remote(self):
        '''
        Set control to remote & unlocked

        Input:
            None

        Output:
            None
        '''
        logging.info(__name__ + ' : Set control to remote & unlocked')
        self.set_remote_status(3)

    def local(self):
        '''
        Set control to local & locked

        Input:
            None

        Output:
            None
        '''
        logging.info(__name__ + ' : Set control to local & locked')
        self.set_remote_status(0)

    def set_remote_status(self, mode):
        '''
        Set remote control status.

        Input:
            mode(int) :
            0 : "Local and locked",
            1 : "Remote and locked",
            2 : "Local and unlocked",
            3 : "Remote and unlocked",

        Output:
            None
        '''
        status = {
        0 : "Local and locked",
        1 : "Remote and locked",
        2 : "Local and unlocked",
        3 : "Remote and unlocked",
        }
        logging.info(__name__ + ' : Setting remote control status to %s' % status.get(mode,"Unknown"))
        self._execute('C%s' % mode)

    def do_get_1K_pot_temp(self):
        '''
        Get 1K Pot Temperature from device.
        Input:
            None

        Output:
            result (float) : 1K Pot Temperature in mK
        '''
        logging.info(__name__ + ' : Read 1K Pot Temperature')
        result = self._execute('R2')
        return float(result.replace('R',''))

    def do_get_mix_chamber_temp(self):
        '''
        Get Mix Chamber Temperature
        Input:
            None

        Output:
            result (float) : Mix Chamber Temperature in mK
        '''
        logging.info(__name__ + ' : Read Mix Chamber Temperature')
        result = self._execute('R3')
        return float(result.replace('R',''))



from qkit.core.instrument_basev2 import QkitProperty, QkitFunction, interval_check, ModernInstrument
        

class Channel:

    def __init__(self, id):
        self.id = id

    @QkitProperty(type=float, units="A", tags=["Current Supply"])
    def amperage(self):
        """
        Provides the amperage of something.
        """
        return 3.14

    @amperage.setter
    def amperage(self, value):
        print(f"Set amp to {value}")

    @QkitFunction
    def do_something(self):
        return f"Maybe{self.id}"

class ModernDriverTest(ModernInstrument):

    def __init__(self, name):
        super().__init__(name)
        print("In init", name)
        self.channels = [Channel(i) for i in range(8)]
        self.discover_capabilities()

    @QkitProperty(type=float, units="V", tags=["Voltage Supply"], arg_checker=interval_check(0.0, 5.0))
    def voltage(self):
        """
        Measures the voltage of something.
        """
        return 5

    @voltage.setter
    def voltage(self, v):
        print(f"Set voltage to {v}")

    @QkitFunction
    def do_complicated_stuff(self):
        return "Nope"


# -*- coding: utf-8 -*-
# Micha Wildermuth, micha.wildermuth@kit.edu 2020

from qkit.core.instrument_base import Instrument


class mfc_dummy(Instrument):
    '''
    This is a driver for a dummy mass flow controller as used for sputter deposition monitoring qkit.services.qdepokit.sdi.

    Usage:
        Initialize with
        <name> = qkit.instruments.create('<name>', 'mfc_dummy')
    '''
    def __init__(self, name):
        self.__name__ = __name__
        Instrument.__init__(self, name, tags=['virtual'])
        self.predef_channels = {'Ar': 0,
                                'ArO': 0,
                                'N': 0,
                                'O': 0}
        self.add_function('get_pressure')
        self.add_function('get_flow')

    def get_pressure(self):
        return 1

    def get_flow(self, channel=1):
        return channel


import numpy as np
import logging
import collections

logging.basicConfig(level=logging.INFO)
log = logging.getLogger("FileParser")


def _parse_parameter_line(line) -> dict[str, float]:
    """
    Takes the line and splits it by spaces. All asignments at the end are taken as parameters.
    """
    parameter_definition = line.split(' ')
    parameter_definition.reverse()
    params = {}
    for candidate in parameter_definition:
        if not '=' in candidate:
            # This is the first non-assignment, starting from the back.
            # This indicates we went through all the assignments and can stop.
            break
        split = candidate.split('=')
        params[split[0]] = float(split[1])
    return params

def _test_parse_parameter_line():
    log.info("Testing _parse_parameter_line")
    test_case = "/home/thilo/nextcloud/Projekt-Promotion/samples/v11-Simulation-update-freq/v11-5-simulation-hot-coupling.son DE_EMBEDDED L=81.5 L2=94.0"
    result = _parse_parameter_line(test_case)
    assert result == {"L": 81.5, "L2": 94.0}, f"Expected {result} to be {{'L': 81.5, 'L2': 94.0}}"

def _parse_data_line(line):
    """
    A data line is a comma separated list of floats.
    """
    return [float(segement) for segement in line.split(',')]

def _test_data_line_parsing():
    log.info("Testing _parse_data_line")
    test_case = "4.82,3.42233e-05"
    result = _parse_data_line(test_case)
    assert result == [4.82, 3.42233e-05], f"Expected {result} to be [4.82, 3.42233e-05]"

def parse_file_lines(lines, optimize_length=False):
    parsing_data = False # We have a simple binary state machine here.
    parameters = []
    # All the data across all parameters
    frequencies = []
    amplitudes = []

    block_frequencies = []
    block_amplitudes = []

    lines_iterator = iter(lines)
    while True:
        try:
            line = next(lines_iterator).strip()
        except StopIteration:
            break
        if not line:
            continue
        if parsing_data:
            try:
                # We are currently parsing data. Try to parse the line as data.
                data = _parse_data_line(line)
                block_frequencies.append(data[0])
                block_amplitudes.append(data[1])
            except ValueError:
                # If parsing as data failed, we are done with the data block.
                parsing_data = False
                # Append to total data
                frequencies.append(block_frequencies)
                amplitudes.append(block_amplitudes)
                log.debug(f"Found data block with {len(block_frequencies)} points")
                # Reset block data
                block_frequencies = []
                block_amplitudes = []
                # Drop down to not parsing data
        
        if not parsing_data:
            header = _parse_parameter_line(line)
            parameters.append(header)
            log.debug(f"Found parameter {header}")
            # Skip the table head
            next(lines_iterator)
            parsing_data = True
    
    # After parsing the file, append the data from the last block.
    frequencies.append(block_frequencies)
    amplitudes.append(block_amplitudes)

    # Sometimes Sonnet is weird and has different lengths for the data. Clean it up.
    if optimize_length:
        # Determine most common line length
        lengths = [len(f) for f in frequencies]
        most_common_length = max(set(lengths), key=lengths.count)
        # Remove all lines that do not have the most common length
        frequencies = [f for f in frequencies if len(f) == most_common_length]
        amplitudes = [a for a in amplitudes if len(a) == most_common_length]
    return np.asarray(parameters), np.asarray(frequencies), np.asarray(amplitudes)

def parse_file(file_path, optimize_length=False):
    with open(file_path, 'r') as file:
        lines = file.readlines()
    return parse_file_lines(lines, optimize_length)

def _test_file_parser():
    log.info("Testing file parser")
    import textwrap
    test_case = textwrap.dedent="""\
    /home/thilo/nextcloud/Projekt-Promotion/samples/v11-Simulation-update-freq/v11-5-simulation-hot-coupling.son DE_EMBEDDED L=90.0
    FREQUENCY (GHz) MAG[S21]
    3,1.58932e-05
    3.01,1.62456e-05
    3.02,1.66094e-05
    3.03,1.69852e-05
    """
    e = None
    try:
        parameters, frequencies, amplitudes = parse_file_lines(test_case.split('\n'))
        assert parameters[0] == {"L": 90.0}, f"Expected {parameters[0]} to be {{'L': 90.0}}"
        assert frequencies[0][0] == 3, f"Expected {frequencies[0][0]} to be 3"
        assert amplitudes[0][0] == 1.58932e-05, f"Expected {amplitudes[0][0]} to be 1.58932e-05"
    except Exception as e:
        log.error(e)
        assert e == None, "Expected no exception"

if __name__ == "__main__":
    _test_parse_parameter_line()
    _test_data_line_parsing()
    _test_file_parser()
    log.info("All tests passed.")


# QTLAB class for remote access of an TIP temperature control server version 2,
# Author: HR @ KIT 2019-

import time
import types
from numpy import arange, size, linspace, sqrt, ones, delete, append, argmin, array, abs
import logging
import zmq
import json
import qkit
from qkit.core.instrument_base import Instrument



class Error(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)

# zero.1 version of a remote tip command


RANGE = {0: 0.02,
         1: 0.2,
         2: 2,
         3: 20,
         4: 200,
         5: 2e3,
         6: 20e3,
         7: 200e3,
         8: 2e6,
         9: 20e6
         }


class tip2_client(Instrument):
    """
        This is the remote tip client to connect to the TIP2 temperature control program

        Usage:
        Initialize with
        <name> = instruments.create('<name>', 'tip2_client', url = "tcp://localhost:5000",setup = True)
        # If setup is True, the controled devices are automatically provided as parameters. Currently only supported for thermometers.
        # If there is anything going wrong with this automatic process, you can set setup=False and create the devices manually.
        # If you want to control the temperature of a device called 'mxc', you can do:
        name.get_mxc_temperature()
        name.enable_PID('mxc')
        name.r_set_T(0.050) # temperature in Kelvin
        
    """

    def __init__(self, name, url = "tcp://localhost:5000",setup = True):
        
        Instrument.__init__(self, name, tags=['physical'])

        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.REQ)
        self.socket.setsockopt(zmq.LINGER, 0)
        self.setup_connection(url=url)
        self.default_device = "mxc"
        self.control_device = ''
        self.T = 0.0

        self.add_function('set_param')
        self.add_function('get_param')
        self.add_function('get_device')
        self.add_function('get_devices')
        self.add_function('get_config')
        self.add_function('get_controled_thermometers')
        self.add_function('define_default_thermometer')
        self.add_function("setup_devices")

        self.add_function('r_get_T')
        self.add_function('r_get_R')
        self.add_function('new_T')
        self.add_function('close')
        
        
        # self.add_parameter('range', type=int,
        #                    flags=Instrument.FLAG_GETSET)
        # self.add_parameter('excitation', type=int,
        #                    flags=Instrument.FLAG_GETSET)
        if setup:
            self.setup_devices()

    def close(self):
        print ("closing zmq socket")
        self.socket.close()

    def setup_connection(self,url="tcp://localhost:5000"):
        print("Connecting to TIP server...")
        self.socket.connect(url)
    
    def setup_devices(self):
        for d in self.get_devices():
            i = self.get_device(d)
            if i['type'] == "thermometer":
                self.add_parameter("%s_temperature"%d,type=float,flags=Instrument.FLAG_GET,
                                   units=i['unit'],get_func=lambda d=d:self.get_param(d,'temperature'))
                self.add_parameter("%s_active"%d, type=bool, flags=Instrument.FLAG_GETSET,
                                   get_func=lambda d=d:self._boolean(self.get_param(d,"active")),
                                   set_func=lambda x,d=d: self.set_param(d,"active",str(x)))
                self.add_parameter("%s_interval" % d, type=float, flags=Instrument.FLAG_GETSET,
                                   get_func=lambda d=d: self.get_param(d, "interval"),
                                   set_func=lambda x, d=d: self.set_param(d, "interval", str(x)))
                self.get(["%s_temperature"%d,"%s_active"%d,"%s_interval" % d])
                if d==self.default_device:
                    self.add_parameter('P', flags=Instrument.FLAG_GETSET, type=float, units='')
                    self.add_parameter('I', flags=Instrument.FLAG_GETSET, type=float, units='')
                    self.add_parameter('D', flags=Instrument.FLAG_GETSET, type=float, units='')
    

    def set_param(self,device, param, value):
        self.socket.send_string("set/"+device+"/"+param+"/"+str(value))
        message = self.socket.recv_string()
        return message

    def get_param(self,device, param):
        self.socket.send_string("get/"+device+"/"+param)
        message = self.socket.recv_string()
        return message

    def get_device(self,device):
        self.socket.send_string("get/"+device+"/:")
        message = self.socket.recv_string()
        return json.loads(message)

    def get_devices(self):
        self.socket.send_string("get/:")
        message = self.socket.recv_string()
        return json.loads(message)

    def get_config(self):
        self.socket.send_string("get/::")
        message = self.socket.recv_string()
        return json.loads(message)

    def get_controled_thermometers(self):
        devices = self.get_devices()
        controlled_devices = []
        for dev in devices:
            if self._boolean(self.get_param(dev,'active')):
                if self._boolean(self.get_param(dev,'control_active')):
                    controlled_devices.append(dev)

        return controlled_devices
    
    def define_default_thermometer(self,thermometer):
        self.default_device = thermometer
    
    def enable_PID(self,channel):
        """
        Enables the PID control for the given channel.
        :param channel: Channel name, i.e. "mxc"
        :return:
        """
        rv = self.set_param(channel,"control_active",True)
        if not self._boolean(rv):
            raise ValueError("enable_PID not successful, device responded '%s'"%rv)
        self.default_device = channel
        
    def disable_PID(self,channel=None):
        if channel is None:
            channel = self.default_device
        self.set_param(channel, "control_active", False)
        return self.get_controled_thermometers()
    
    # get and set Temperature

    def r_set_T(self, T, thermometer = None, safety_checks = True):
        if thermometer is None:
            thermometer = self.default_device
        self.T = T
        if T > 1.0 and safety_checks:
            raise  ValueError(__name__+": r_set_T cancelled. Target temperature > 1K. Make sure what you are doing and disable safety checks.")
        self.set_param(thermometer,'control_temperature',T)

    def r_get_T(self,thermometer = None):
        if thermometer is None:
            thermometer = self.default_device
        return float(self.get_param(thermometer,'temperature'))

    def r_get_R(self, thermometer = None):
        if thermometer is None: thermometer = self.default_device
        return float(self.get_param(thermometer,'resistance'))

    def new_T(self, T, dT_max=0.0005, thermometer = None):
        if thermometer is None: thermometer = self.default_device
        def rms(Ts):
            return sqrt(sum(Ts * Ts) / len(Ts))

        Ts = ones(20)
        settling_time = time.time()
        print ("T set to "+str(T))
        self.r_set_T(T)

        T_current = self.r_get_T()
        print (T_current)
        while True:
            T_current = self.r_get_T()
            Ts = delete(append(Ts, T_current), 0)
            rmsTs = rms(Ts)

            if abs(rmsTs - T) > dT_max:
                print ("dT > dT_max(%.5f): %.5f at Tctl: %.5f Curr T: %.5f" % (dT_max, rmsTs - T, T, T_current))
                qkit.flow.sleep(2)
            else:
                break
        print ("settling time: %s"%(time.time() - settling_time))


    def do_set_T(self, val):
        try:
            self.r_set_T(val)
            self.T = self.r_get_T()
            return self.T
        except ValueError:
            logging.warning('TIP connection probably lost. Nothing set.')
            return False

    def do_get_T(self,thermometer = None ):
        if thermometer is None: thermometer = self.default_device
        return float(self.get_param(thermometer,'temperature'))

    def do_get_P(self,thermometer = None):
        if thermometer is None: thermometer = self.default_device
        return float(self.get_param(thermometer,'control_P'))

    def do_set_P(self, P,thermometer = None):
        if thermometer is None: thermometer = self.default_device
        return float(self.set_param(thermometer,'control_P',P))

    def do_get_I(self,thermometer = None): 
        if thermometer is None: thermometer = self.default_device       
        return float(self.get_param(thermometer,'control_I'))

    def do_set_I(self, I,thermometer = None):
        if thermometer is None: thermometer = self.default_device
        return float(self.set_param(thermometer,'control_I',I))

    def do_get_D(self,thermometer = None):
        if thermometer is None: thermometer = self.default_device
        return float(self.get_param(thermometer,'control_D'))

    def do_set_D(self, D,thermometer = None):
        if thermometer is None: thermometer = self.default_device
        return float(self.set_param(thermometer,'control_D',D))

    # bridge settings for different channels

    def do_get_range(self, thermometer = None):
        if thermometer is None: thermometer = self.default_device
        return float(self.get_param(thermometer,'device_range'))

    def do_set_range(self, range, thermometer = None):
        """
        Set the resistance range of the specified channel. Check RANGE dict for help.
        """
        if thermometer is None: thermometer = self.default_device
        self.send("SET/T/%i/Range/%i" % (thermometer, range))
        return bool(self.recv())

    def do_get_excitation(self, thermometer = None):
        if thermometer is None: thermometer = self.default_device
        return float(self.get_param(thermometer,'device_excitation'))

    def do_set_excitation(self, excitation, thermometer = None):
        """
        set the measurement excitation of the specified channel.
        -1: Excitation off
        -1: (excitation off)
        0:3uV
        1:10uV
        2:30uV
        3:100uV
        4:300uV
        5:1 mV
        6:3 mV
        7:10 mV
        8:30 mV
        """
        if thermometer is None: thermometer = self.default_device
        self.send("SET/T/%i/EX/%i" % (thermometer, excitation))
        return bool(self.recv())
    
    def do_get_temperature(self, thermometer = None):
        if thermometer is None: thermometer = self.default_device
        return float(self.get_param(thermometer,'temperature'))
    

    def do_get_resistance(self, thermometer = None):
        if thermometer is None: thermometer = self.default_device
        return float(self.get_param(thermometer,'resistance'))

    def _boolean(self,s): return s.lower() in ("yes", "true", "t", "1")



# edw_client.py by JB@KIT 03/2015
# test client for edw server

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import socket
import sys

HOST, PORT = 'ip-address', 9955
data = 'jochen'

# Create a socket (SOCK_STREAM means a TCP socket)
sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

try:
    # Connect to server and send data
    sock.connect((HOST, PORT))
    sock.sendall(data + "\n")

    # Receive data from the server and shut down
    received = sock.recv(1024)
finally:
    sock.close()

print "Sent:     {}".format(data)
print "Received: {}".format(received)

